import operator
import json
import numpy as np

from pandas.core.arrays import ExtensionArray
from pandas.core.dtypes.dtypes import ExtensionDtype
from arcgis.geometry import Geometry
import pandas as pd
from .parser import _to_geo_array



class NumPyBackedExtensionArrayMixin(ExtensionArray):
    """
    Geo-Specific Extension Array Mixin
    """
    @property
    def dtype(self):
        """The dtype for this extension array, GeoType"""
        return self._dtype

    @classmethod
    def _from_sequence(cls, scalars):
        return cls(scalars)

    @classmethod
    def _constructor_from_sequence(cls, scalars):
        return cls(scalars)

    @classmethod
    def _from_factorized(cls, values, original):
        return cls(values)

    @property
    def shape(self):
        return (len(self.data),)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, *args):
        result = operator.getitem(self.data, *args)
        if isinstance(result, (dict, Geometry)):
            return result
        elif isinstance(result, type(None)) or \
             isinstance(result, type(np.nan)):
            return None
        elif not isinstance(result, GeoArray):
            return GeoArray(result)
        return result


    def setitem(self, indexer, value):
        """Set the 'value' inplace.
        """

        self[indexer] = value
        return self

    @property
    def nbytes(self):
        return self._itemsize * len(self)

    def _formatting_values(self):
        return np.array(self._format_values(), dtype='object')

    def copy(self, deep=False):
        return type(self)(self.data.copy())

    @classmethod
    def _concat_same_type(cls, to_concat):
        return cls(np.concatenate([array.data for array in to_concat]))

    def tolist(self):
        return self.data.tolist()

    def argsort(self, axis=-1, kind='quicksort', order=None):
        return self.data.argsort()

    def unique(self):
        _, indices = np.unique(self.data, return_index=True)
        data = self.data.take(np.sort(indices))
        return self._from_ndarray(data)

class GeoType(ExtensionDtype):
    name = 'geometry'
    type = Geometry
    kind = 'O'
    _record_type = np.dtype('O')
    na_value = None

    @classmethod
    def construct_from_string(cls, string):
        if string == cls.name:
            return cls()
        else:
            raise TypeError("Cannot construct a '{}' from "
                            "'{}'".format(cls, string))

class GeoArray(NumPyBackedExtensionArrayMixin):
    """Array for Geometry data.
    """
    _dtype = GeoType()
    _itemsize = 8
    ndim = 1
    can_hold_na = True

    def __init__(self, values, copy=True):
        self.data = np.array(values, dtype='O', copy=copy)

    @classmethod
    def _from_ndarray(cls, data, copy=False):
        return cls(data, copy=copy)

    @property
    def na_value(self):
        return self.dtype.na_value

    def __repr__(self):
        formatted = self._format_values()
        return "GeoArray({!r})".format(formatted)

    def __str__(self):
        return self.__repr__()

    def _format_values(self):
        if self.data.ndim == 0:
            return ""
        return [_format(x) if x else None for x in self.data]

    @classmethod
    def from_geometry(cls, data, copy=False):
        """"""
        if copy:
            data = data.copy()
        new = GeoArray([])
        new.data = np.array(data)
        return new

    def __setitem__(self, key, value):
        if value is None or  \
           (isinstance(value, str) and value == ""):
            self.data[key] = value
        else:
            value = Geometry(value)
            self.data[key] = value

    def __iter__(self):
        return iter(self.data.tolist())

    def __eq__(self, other):
        return self.data == other

    def equals(self, other):
        if not isinstance(other, type(self)):
            raise TypeError
        return (self.data == other.data).all()

    def _values_for_factorize(self):
        # Should hit pandas' UInt64Hashtable
        return self, 0

    def isna(self):
        return (self.data == self._dtype.na_value)

    @property
    def _parser(self):
        return lambda x: x

    def take(self, indexer, allow_fill=True, fill_value=None):
        mask = indexer == -1
        result = self.data.take(indexer)
        result[mask] = self.dtype.na_value
        return type(self)(result, copy=False)

    def _formatting_values(self):
        return np.array(self._format_values(), dtype='object')

    @classmethod
    def _concat_same_type(cls, to_concat):
        return cls(np.concatenate([array.data for array in to_concat]))

    def take_nd(self, indexer, allow_fill=True, fill_value=None):
        return self.take(indexer, allow_fill=allow_fill, fill_value=fill_value)

    def copy(self, deep=False):
        return type(self)(self.data.copy())

    from arcgis.geometry import BaseGeometry
    @property
    def is_valid(self):
        """Checks if the Geometry is Valid"""
        return pd.Series([g.is_valid for g in self])

    #----------------------------------------------------------------------
    def _call_property(self, prop, as_ga=False):
        """accesses a property on a dataframe"""
        fn = lambda g: getattr(g, prop, None)
        vals = np.vectorize(fn,otypes='O')(self.data)
        if as_ga:
            s = pd.Series(GeoArray(values=vals))
        else:
            s = pd.Series(vals)
        s.name = prop
        return s
    #----------------------------------------------------------------------
    def _call_method(self, name, is_ga=False, **kwargs):
        """accesses a method on a dataframe"""
        import warnings
        warnings.simplefilter("ignore")
        fn = lambda g, n: getattr(g, n, None)(**kwargs) \
            if g is not None else None
        vals = np.vectorize(fn, otypes='O')(self.data, name)
        if is_ga:
            s = pd.Series(GeoArray(values=vals, copy=False))
        else:
            s = pd.Series(vals)
        s.name = name
        return s
    #----------------------------------------------------------------------
    @property
    def area(self):
        """returns the geometry area"""
        return self._call_property('area')
    #----------------------------------------------------------------------
    @property
    def as_arcpy(self):
        """returns the geometry area"""
        return self._call_property('as_arcpy')
    #----------------------------------------------------------------------
    @property
    def as_shapely(self):
        """returns the geometry area"""
        return self._call_property('as_shapely')
    #----------------------------------------------------------------------
    @property
    def centroid(self):
        """returns Geometry centroid"""
        return self._call_property('centroid')
    #----------------------------------------------------------------------
    @property
    def extent(self):
        """returns the extent of the geometry"""
        return self._call_property("extent")
    #----------------------------------------------------------------------
    @property
    def first_point(self):
        """
        The first coordinate point of the geometry for each entry.
        """
        return self._call_property("first_point", as_ga=True)
    #----------------------------------------------------------------------
    @property
    def geoextent(self):
        return self._call_property("geoextent")
    #----------------------------------------------------------------------
    @property
    def geometry_type(self):
        return self._call_property("geometry_type")
    #----------------------------------------------------------------------
    @property
    def hull_rectangle(self):
        return self._call_property("hull_rectangle")
    #----------------------------------------------------------------------
    @property
    def is_empty(self):
        return self._call_property("is_empty")
    #----------------------------------------------------------------------
    @property
    def is_multipart(self):
        return self._call_property("is_multipart")
    #----------------------------------------------------------------------
    @property
    def is_valid(self):
        return self._call_property("is_valid")
    #----------------------------------------------------------------------
    @property
    def JSON(self):
        return self._call_property("JSON")
    #----------------------------------------------------------------------
    @property
    def label_point(self):
        return self._call_property("label_point", as_ga=True)
    #----------------------------------------------------------------------
    @property
    def last_point(self):
        return self._call_property("last_point", as_ga=True)
    #----------------------------------------------------------------------
    @property
    def length(self):
        return self._call_property("length", as_ga=False)
    #----------------------------------------------------------------------
    @property
    def length3D(self):
        return self._call_property("length3D", as_ga=False)
    #----------------------------------------------------------------------
    @property
    def part_count(self):
        return self._call_property("part_count")
    #----------------------------------------------------------------------
    @property
    def point_count(self):
        return self._call_property("point_count")
    #----------------------------------------------------------------------
    @property
    def spatial_reference(self):
        return self._call_property("spatial_reference")
    #----------------------------------------------------------------------
    @property
    def true_centroid(self):
        return self._call_property("true_centroid", as_ga=True)
    #----------------------------------------------------------------------
    @property
    def WKB(self):
        return self._call_property("WKB")
    #----------------------------------------------------------------------
    @property
    def WKT(self):
        return self._call_property("WKT")
    #----------------------------------------------------------------------
    def angle_distance_to(self, second_geometry, method="GEODESIC"):
        """
        Returns a tuple of angle and distance to another point using a
        measurement type.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required Geometry.  A arcgis.Geometry object.
        ---------------     --------------------------------------------------------------------
        method              Optional String. PLANAR measurements reflect the projection of geographic
                            data onto the 2D surface (in other words, they will not take into
                            account the curvature of the earth). GEODESIC, GREAT_ELLIPTIC,
                            LOXODROME, and PRESERVE_SHAPE measurement types may be chosen as
                            an alternative, if desired.
        ===============     ====================================================================

        :returns: a tuple of angle and distance to another point using a measurement type.
        """
        return self._call_method(name='angle_distance_to',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry,
                                    'method' : method})
    #----------------------------------------------------------------------
    def boundary(self):
        """
        Constructs the boundary of the geometry.

        :returns: arcgis.geometry.Polyline
        """

        return self._call_method(name='boundary', is_ga=True)
    #----------------------------------------------------------------------
    def buffer(self, distance):
        """
        Constructs a polygon at a specified distance from the geometry.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        distance            Required float. The buffer distance. The buffer distance is in the
                            same units as the geometry that is being buffered.
                            A negative distance can only be specified against a polygon geometry.
        ===============     ====================================================================

        :returns: arcgis.geometry.Polygon
        """
        return self._call_method(name='buffer',
                                 is_ga=True,
                                 **{'distance' : distance})
    #----------------------------------------------------------------------
    def clip(self, envelope):
        """
        Constructs the intersection of the geometry and the specified extent.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        envelope            required tuple. The tuple must have (XMin, YMin, XMax, YMax) each value
                            represents the lower left bound and upper right bound of the extent.
        ===============     ====================================================================

        :returns: output geometry clipped to extent

        """
        return self._call_method(name='clip',
                                 is_ga=True,
                                 **{'envelope' : envelope})
    #----------------------------------------------------------------------
    def contains(self, second_geometry, relation=None):
        """
        Indicates if the base geometry contains the comparison geometry.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        relation            Optional string. The spatial relationship type.

                            + BOUNDARY - Relationship has no restrictions for interiors or boundaries.
                            + CLEMENTINI - Interiors of geometries must intersect. Specifying CLEMENTINI is equivalent to specifying None. This is the default.
                            + PROPER - Boundaries of geometries must not intersect.
        ===============     ====================================================================

        :returns: boolean
        """
        return self._call_method(name='contains',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry,
                                    'relation' : relation})
    #----------------------------------------------------------------------
    def convex_hull(self):
        """
        Constructs the geometry that is the minimal bounding polygon such
        that all outer angles are convex.
        """
        return self._call_method(name='convex_hull',
                                 is_ga=True)
    #----------------------------------------------------------------------
    def crosses(self, second_geometry):
        """
        Indicates if the two geometries intersect in a geometry of a lesser
        shape type.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :returns: boolean

        """
        return self._call_method(name='crosses',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry})
    #----------------------------------------------------------------------
    def cut(self, cutter):
        """
        Splits this geometry into a part left of the cutting polyline, and
        a part right of it.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        cutter              Required Polyline. The cuttin polyline geometry
        ===============     ====================================================================

        :returns: a list of two geometries

        """
        return self._call_method(name='cut',
                                 is_ga=True,
                                 **{'cutter' : cutter})
    #----------------------------------------------------------------------
    def densify(self, method, distance, deviation):
        """
        Creates a new geometry with added vertices

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        method              Required String. The type of densification, DISTANCE, ANGLE, or GEODESIC
        ---------------     --------------------------------------------------------------------
        distance            Required float. The maximum distance between vertices. The actual
                            distance between vertices will usually be less than the maximum
                            distance as new vertices will be evenly distributed along the
                            original segment. If using a type of DISTANCE or ANGLE, the
                            distance is measured in the units of the geometry's spatial
                            reference. If using a type of GEODESIC, the distance is measured
                            in meters.
        ---------------     --------------------------------------------------------------------
        deviation           Required float. Densify uses straight lines to approximate curves.
                            You use deviation to control the accuracy of this approximation.
                            The deviation is the maximum distance between the new segment and
                            the original curve. The smaller its value, the more segments will
                            be required to approximate the curve.
        ===============     ====================================================================

        :returns: arcgis.geometry.Geometry

        """
        return self._call_method(name='densify',
                                 is_ga=True,
                                 **{'method' : method,
                                    'distance' : distance,
                                    'deviation' : deviation})
    #----------------------------------------------------------------------
    def difference(self, second_geometry):
        """
        Constructs the geometry that is composed only of the region unique
        to the base geometry but not part of the other geometry. The
        following illustration shows the results when the red polygon is the
        source geometry.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :returns: arcgis.geometry.Geometry

        """
        return self._call_method(name='difference',
                                 is_ga=True,
                                 **{'second_geometry' : second_geometry})
    #----------------------------------------------------------------------
    def disjoint(self, second_geometry):
        """
        Indicates if the base and comparison geometries share no points in
        common.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :returns: boolean

        """
        return self._call_method(name='disjoint',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry})
    #----------------------------------------------------------------------
    def distance_to(self, second_geometry):
        """
        Returns the minimum distance between two geometries. If the
        geometries intersect, the minimum distance is 0.
        Both geometries must have the same projection.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :returns: float

        """
        return self._call_method(name='distance_to',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry})
    #----------------------------------------------------------------------
    def equals(self, second_geometry):
        """
        Indicates if the base and comparison geometries are of the same
        shape type and define the same set of points in the plane. This is
        a 2D comparison only; M and Z values are ignored.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :returns: boolean


        """
        return self._call_method(name='equals',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry})
    #----------------------------------------------------------------------
    def generalize(self, max_offset):
        """
        Creates a new simplified geometry using a specified maximum offset
        tolerance.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        max_offset          Required float. The maximum offset tolerance.
        ===============     ====================================================================

        :returns: arcgis.geometry.Geometry

        """
        return self._call_method(name='generalize',
                                 is_ga=True,
                                 **{'max_offset' : max_offset})
    #----------------------------------------------------------------------
    def get_area(self, method, units=None):
        """
        Returns the area of the feature using a measurement type.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        method              Required String. LANAR measurements reflect the projection of
                            geographic data onto the 2D surface (in other words, they will not
                            take into account the curvature of the earth). GEODESIC,
                            GREAT_ELLIPTIC, LOXODROME, and PRESERVE_SHAPE measurement types
                            may be chosen as an alternative, if desired.
        ---------------     --------------------------------------------------------------------
        units               Optional String. Areal unit of measure keywords: ACRES | ARES | HECTARES
                            | SQUARECENTIMETERS | SQUAREDECIMETERS | SQUAREINCHES | SQUAREFEET
                            | SQUAREKILOMETERS | SQUAREMETERS | SQUAREMILES |
                            SQUAREMILLIMETERS | SQUAREYARDS
        ===============     ====================================================================

        :returns: float

        """
        return self._call_method(name='get_area',
                                 is_ga=False,
                                 **{'method' : method,
                                    'units' : units})
    #----------------------------------------------------------------------
    def get_length(self, method, units):
        """
        Returns the length of the feature using a measurement type.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        method              Required String. PLANAR measurements reflect the projection of
                            geographic data onto the 2D surface (in other words, they will not
                            take into account the curvature of the earth). GEODESIC,
                            GREAT_ELLIPTIC, LOXODROME, and PRESERVE_SHAPE measurement types
                            may be chosen as an alternative, if desired.
        ---------------     --------------------------------------------------------------------
        units               Required String. Linear unit of measure keywords: CENTIMETERS |
                            DECIMETERS | FEET | INCHES | KILOMETERS | METERS | MILES |
                            MILLIMETERS | NAUTICALMILES | YARDS
        ===============     ====================================================================

        :returns: float

        """
        return self._call_method(name='get_length',
                                 is_ga=False,
                                 **{'method' : method,
                                    'units' : units})
    #----------------------------------------------------------------------
    def get_part(self, index=None):
        """
        Returns an array of point objects for a particular part of geometry
        or an array containing a number of arrays, one for each part.

        **requires arcpy**

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        index               Required Integer. The index position of the geometry.
        ===============     ====================================================================

        :return: arcpy.Array

        """
        return self._call_method(name='get_part',
                                 is_ga=False,
                                 **{'index' : index})
    #----------------------------------------------------------------------
    def intersect(self, second_geometry, dimension=1):
        """
        Constructs a geometry that is the geometric intersection of the two
        input geometries. Different dimension values can be used to create
        different shape types. The intersection of two geometries of the
        same shape type is a geometry containing only the regions of overlap
        between the original geometries.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        dimension           Required Integer. The topological dimension (shape type) of the
                            resulting geometry.

                            + 1  -A zero-dimensional geometry (point or multipoint).
                            + 2  -A one-dimensional geometry (polyline).
                            + 4  -A two-dimensional geometry (polygon).

        ===============     ====================================================================

        :returns: boolean

        """
        return self._call_method(name='intersect',
                                 is_ga=True,
                                 **{'second_geometry' : second_geometry,
                                    'dimension' : dimension})
    #----------------------------------------------------------------------
    def measure_on_line(self, second_geometry, as_percentage=False):
        """
        Returns a measure from the start point of this line to the in_point.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        as_percentage       Optional Boolean. If False, the measure will be returned as a
                            distance; if True, the measure will be returned as a percentage.
        ===============     ====================================================================

        :return: float

        """
        return self._call_method(name='measure_on_line',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry,
                                    'as_percentage' : as_percentage})
    #----------------------------------------------------------------------
    def overlaps(self, second_geometry):
        """
        Indicates if the intersection of the two geometries has the same
        shape type as one of the input geometries and is not equivalent to
        either of the input geometries.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: boolean

        """
        return self._call_method(name='overlaps',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry})
    #----------------------------------------------------------------------
    def point_from_angle_and_distance(self, angle, distance, method='GEODESCIC'):
        """
        Returns a point at a given angle and distance in degrees and meters
        using the specified measurement type.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        angle               Required Float. The angle in degrees to the returned point.
        ---------------     --------------------------------------------------------------------
        distance            Required Float. The distance in meters to the returned point.
        ---------------     --------------------------------------------------------------------
        method              Optional String. PLANAR measurements reflect the projection of geographic
                            data onto the 2D surface (in other words, they will not take into
                            account the curvature of the earth). GEODESIC, GREAT_ELLIPTIC,
                            LOXODROME, and PRESERVE_SHAPE measurement types may be chosen as
                            an alternative, if desired.
        ===============     ====================================================================

        :return: arcgis.geometry.Geometry


        """
        return self._call_method(name='point_from_angle_and_distance',
                                 is_ga=True,
                                 **{'angle' : angle,
                                    'distance' : distance,
                                    'method' : method})
    #----------------------------------------------------------------------
    def position_along_line(self, value, use_percentage=False):
        """
        Returns a point on a line at a specified distance from the beginning
        of the line.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        value               Required Float. The distance along the line.
        ---------------     --------------------------------------------------------------------
        use_percentage      Optional Boolean. The distance may be specified as a fixed unit
                            of measure or a ratio of the length of the line. If True, value
                            is used as a percentage; if False, value is used as a distance.
                            For percentages, the value should be expressed as a double from
                            0.0 (0%) to 1.0 (100%).
        ===============     ====================================================================

        :return: arcgis.gis.Geometry

        """
        return self._call_method(name='position_along_line',
                                 is_ga=True,
                                 **{'value' : value,
                                    'use_percentage' : use_percentage})
    #----------------------------------------------------------------------
    def project_as(self, spatial_reference, transformation_name=None):
        """
        Projects a geometry and optionally applies a geotransformation.

        ====================     ====================================================================
        **Argument**             **Description**
        --------------------     --------------------------------------------------------------------
        spatial_reference        Required SpatialReference. The new spatial reference. This can be a
                                 SpatialReference object or the coordinate system name.
        --------------------     --------------------------------------------------------------------
        transformation_name      Required String. The geotransformation name.
        ====================     ====================================================================

        :returns: arcgis.geometry.Geometry
        """
        return self._call_method(name='project_as',
                                 is_ga=True,
                                 **{'spatial_reference' : spatial_reference,
                                    'transformation_name' : transformation_name}
                                 )
    #----------------------------------------------------------------------
    def query_point_and_distance(self, second_geometry,
                                 use_percentage=False):
        """
        Finds the point on the polyline nearest to the in_point and the
        distance between those points. Also returns information about the
        side of the line the in_point is on as well as the distance along
        the line where the nearest point occurs.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        as_percentage       Optional boolean - if False, the measure will be returned as
                            distance, True, measure will be a percentage
        ===============     ====================================================================

        :return: tuple

        """
        return self._call_method(name='query_point_and_distance',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry,
                                    'use_percentage' : use_percentage})
    #----------------------------------------------------------------------
    def segment_along_line(self, start_measure,
                           end_measure, use_percentage=False):
        """
        Returns a Polyline between start and end measures. Similar to
        Polyline.positionAlongLine but will return a polyline segment between
        two points on the polyline instead of a single point.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        start_measure       Required Float. The starting distance from the beginning of the line.
        ---------------     --------------------------------------------------------------------
        end_measure         Required Float. The ending distance from the beginning of the line.
        ---------------     --------------------------------------------------------------------
        use_percentage      Optional Boolean. The start and end measures may be specified as
                            fixed units or as a ratio.
                            If True, start_measure and end_measure are used as a percentage; if
                            False, start_measure and end_measure are used as a distance. For
                            percentages, the measures should be expressed as a double from 0.0
                            (0 percent) to 1.0 (100 percent).
        ===============     ====================================================================

        :returns: Geometry

        """
        return self._call_method(name='segment_along_line',
                                 is_ga=True,
                                 **{'start_measure' : start_measure,
                                    'end_measure' : end_measure,
                                    'use_percentage' : use_percentage})
    #----------------------------------------------------------------------
    def snap_to_line(self, second_geometry):
        """
        Returns a new point based on in_point snapped to this geometry.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: arcgis.gis.Geometry

        """
        return self._call_method(name='snap_to_line',
                                 is_ga=True,
                                 **{'second_geometry' : second_geometry})
    #----------------------------------------------------------------------
    def symmetric_difference (self, second_geometry):
        """
        Constructs the geometry that is the union of two geometries minus the
        instersection of those geometries.

        The two input geometries must be the same shape type.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: arcgis.gis.Geometry
        """
        return self._call_method(name='symmetric_difference',
                                 is_ga=True,
                                 **{'second_geometry' : second_geometry})
    #----------------------------------------------------------------------
    def touches(self, second_geometry):
        """
        Indicates if the boundaries of the geometries intersect.


        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: boolean
        """
        return self._call_method(name='touches',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry})
    #----------------------------------------------------------------------
    def union(self, second_geometry):
        """
        Constructs the geometry that is the set-theoretic union of the input
        geometries.


        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ===============     ====================================================================

        :return: arcgis.gis.Geometry
        """
        return self._call_method(name='union',
                                 is_ga=True,
                                 **{'second_geometry' : second_geometry})
    #----------------------------------------------------------------------
    def within(self, second_geometry, relation=None):
        """
        Indicates if the base geometry is within the comparison geometry.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        second_geometry     Required arcgis.geometry.Geometry. A second geometry
        ---------------     --------------------------------------------------------------------
        relation            Optional String. The spatial relationship type.

                            - BOUNDARY  - Relationship has no restrictions for interiors or boundaries.
                            - CLEMENTINI  - Interiors of geometries must intersect. Specifying CLEMENTINI is equivalent to specifying None. This is the default.
                            - PROPER  - Boundaries of geometries must not intersect.

        ===============     ====================================================================

        :return: boolean

        """
        return self._call_method(name='within',
                                 is_ga=False,
                                 **{'second_geometry' : second_geometry,
                                    'relation' : relation})

def _format(g):
    if g in {None, np.nan}:
        return ""
    return json.dumps(g)




"""
IO operations for Feature Classes
"""
import os
import sys
import shutil
import datetime

import numpy as np
import pandas as pd

try:
    import arcpy
    from arcpy import da
    HASARCPY = True
except:
    HASARCPY = False

try:
    import fiona
    HASFIONA = True
except:
    HASFIONA = False

try:
    import shapefile
    HASPYSHP = True
except:
    HASPYSHP = False
#--------------------------------------------------------------------------
def _infer_type(df, col):
    """
    internal function used to get the datatypes for the feature class if
    the dataframe's _field_reference is NULL or there is a column that does
    not have a dtype assigned to it.

    Input:
     dataframe - spatialdataframe object
    Ouput:
      field type name
    """
    import six
    nn = df[col].notnull()
    nn = list(df[nn].index)
    if len(nn) > 0:
        val = df[col][nn[0]]
        if isinstance(val, six.string_types):
            return "TEXT"
        elif isinstance(val, tuple(list(six.integer_types) + [np.int32])):
            return "INTEGER"
        elif isinstance(val, (float, np.int64 )):
            return "FLOAT"
        elif isinstance(val, datetime):
            return "DATE"
    return "TEXT"
#--------------------------------------------------------------------------
def _geojson_to_esrijson(geojson):
    """converts the geojson spec to esri json spec"""
    if geojson['type'] in ['Polygon', 'MultiPolygon']:
        return {
            'rings' : geojson['coordinates'],
            'spatialReference' : {'wkid' : 4326}
        }
    elif geojson['type'] == "Point":
        return {
            "x" : geojson['coordinates'][0],
            "y" : geojson['coordinates'][1],
            'spatialReference' : {'wkid' : 4326}
        }
    elif geojson['type'] == "MultiPoint":
        return {
            "points" : geojson['coordinates'],
            'spatialReference' : {'wkid' : 4326}
        }
    elif geojson['type'] in ['LineString', 'MultiLineString']:
        return {
            "paths" : geojson['coordinates'],
            'spatialReference' : {'wkid' : 4326}
        }
    return geojson
#--------------------------------------------------------------------------
def _geometry_to_geojson(geom):
    """converts the esri json spec to geojson"""
    if 'rings' in geom and \
       len(geom['rings']) == 1:
        return {
            'type' : "Polygon",
            "coordinates" : geom['rings']
        }
    elif 'rings' in geom and \
       len(geom['rings']) > 1:
        return {
            'type' : "MultiPolygon",
            "coordinates" : geom['rings']
        }
    elif geom['type'] == "Point":
        return {
            "coordinates" : [geom['x'], geom['y']],
            "type" : "Point"
        }
    elif geom['type'] == "MultiPoint":
        return {
            "coordinates" : geom['points'],
            'type' : "MultiPoint"
        }
    elif geom['type'].lower() == "polyline" and \
         len(geom['paths']) <= 1:
        return {
            "coordinates" : geom['paths'],
            'type' : "LineString"
        }
    elif geom['type'].lower() == "polyline" and \
         len(geom['paths']) > 1:
        return {
            "coordinates" : geom['paths'],
            'type' : "MultiLineString"
        }
    return geom
#--------------------------------------------------------------------------
def _from_xy(df, x_column, y_column, sr=None):
    """
    Takes an X/Y Column and Creates a Point Geometry from it.
    """
    from arcgis.geometry import SpatialReference, Geometry
    if sr is None:
        sr = SpatialReference({'wkid' : 4326})
    if not isinstance(sr, SpatialReference):
        if isinstance(sr, dict):
            sr = SpatialReference(sr)
        elif isinstance(sr, int):
            sr = SpatialReference({'wkid' : sr})
        elif isinstance(sr, str):
            sr = SpatialReference({'wkt' : sr})
    geoms = []
    for idx, row in df.iterrows():
        geoms.append(
            Geometry({'x' : row[x_column], 'y' : row[y_column],
             'spatialReference' : sr})
        )
    df['SHAPE'] = geoms
    df.spatial.set_geometry('SHAPE')
    return df

#--------------------------------------------------------------------------
def from_table(filename, **kwargs):
    """
    Allows a user to read from a non-spatial table

    **Note: ArcPy is Required for this method**

    ===============     ====================================================
    **Argument**        **Description**
    ---------------     ----------------------------------------------------
    filename            Required string. The path to the table.
    ===============     ====================================================

    **Keyword Arguments**

    ===============     ====================================================
    **Argument**        **Description**
    ---------------     ----------------------------------------------------
    fields              Optional List/Tuple. A list (or tuple) of field
                        names. For a single field, you can use a string
                        instead of a list of strings.

                        Use an asterisk (*) instead of a list of fields if
                        you want to access all fields from the input table
                        (raster and BLOB fields are excluded). However, for
                        faster performance and reliable field order, it is
                        recommended that the list of fields be narrowed to
                        only those that are actually needed.

                        Geometry, raster, and BLOB fields are not supported.

    ---------------     ----------------------------------------------------
    where               Optional String. An optional expression that limits
                        the records returned.
    ---------------     ----------------------------------------------------
    skip_nulls          Optional Boolean. This controls whether records
                        using nulls are skipped.
    ---------------     ----------------------------------------------------
    null_value          Optional String/Integer/Float. Replaces null values
                        from the input with a new value.
    ===============     ====================================================

    :returns: pd.DataFrame

    """
    if HASARCPY:
        where = kwargs.pop("where", None)
        fields = kwargs.pop('fields', "*")
        skip_nulls = kwargs.pop('skip_nulls', True)
        null_value = kwargs.pop("null_value", None)
        return pd.DataFrame(da.TableToNumPyArray(in_table=filename,
                                                 field_names=fields,
                                                 where_clause=where,
                                                 skip_nulls=skip_nulls,
                                                 null_value=null_value))
    elif filename.lower().find('.csv') > -1:
        return pd.read_csv(filename)

    return
#--------------------------------------------------------------------------
def to_table(geo, location, overwrite=True):
    """
    Exports a geo enabled dataframe to a table.

    ===========================     ====================================================================
    **Argument**                    **Description**
    ---------------------------     --------------------------------------------------------------------
    location                        Required string. The output of the table.
    ---------------------------     --------------------------------------------------------------------
    overwrite                       Optional Boolean.  If True and if the table exists, it will be
                                    deleted and overwritten.  This is default.  If False, the table and
                                    the table exists, and exception will be raised.
    ===========================     ====================================================================

    :returns: String
    """
    out_location= os.path.dirname(location)
    fc_name = os.path.basename(location)
    df = geo._data
    if location.lower().find('.csv') > -1:
        geo._df.to_csv(location)
        return location
    elif HASARCPY:
        columns = df.columns.tolist()
        join_dummy = "AEIOUYAJC81Z"
        try:
            columns.pop(columns.index(df.spatial.name))
        except:
            pass
        dtypes = [(join_dummy, np.int64)]
        if overwrite and arcpy.Exists(location):
            arcpy.Delete_management(location)
        elif overwrite == False and arcpy.Exists(location):
            raise ValueError(('overwrite set to False, Cannot '
                              'overwrite the table. '))
        fc = arcpy.CreateTable_management(out_path=out_location,
                                          out_name=fc_name)[0]
        # 2. Add the Fields and Data Types
        #
        oidfld = da.Describe(fc)['OIDFieldName']
        for col in columns[:]:
            if col.lower() in ['fid', 'oid', 'objectid']:
                dtypes.append((col, np.int32))
            elif df[col].dtype.name == 'datetime64[ns]':
                dtypes.append((col, '<M8[us]'))
            elif df[col].dtype.name == 'object':
                try:
                    u = type(df[col][df[col].first_valid_index()])
                except:
                    u = pd.unique(df[col].apply(type)).tolist()[0]
                if issubclass(u, str):
                    mlen = df[col].str.len().max()
                    dtypes.append((col, '<U%s' % int(mlen)))
                else:
                    try:
                        dtypes.append((col, type(df[col][0])))
                    except:
                        dtypes.append((col, '<U254'))
            elif df[col].dtype.name == 'int64':
                dtypes.append((col, np.int64))
            elif df[col].dtype.name == 'bool':
                dtypes.append((col, np.int32))
            else:
                dtypes.append((col, df[col].dtype.type))

        array = np.array([],
                        np.dtype(dtypes))
        arcpy.da.ExtendTable(fc,
                             oidfld, array,
                             join_dummy, append_only=False)
        # 3. Insert the Data
        #
        fields = arcpy.ListFields(fc)
        icols = [fld.name for fld in fields \
                 if fld.type not in ['OID', 'Geometry'] and \
                 fld.name in df.columns]
        dfcols = [fld.name for fld in fields \
                  if fld.type not in ['OID', 'Geometry'] and\
                  fld.name in df.columns]
        with da.InsertCursor(fc, icols) as irows:
            for idx, row in df[dfcols].iterrows():
                try:
                    irows.insertRow(row.tolist())
                except:
                    print("row %s could not be inserted." % idx)
        return fc

    return
#--------------------------------------------------------------------------
def from_featureclass(filename, **kwargs):
    """
    Returns a GeoDataFrame from a feature class.
    Inputs:
     filename: full path to the feature class
    Optional Parameters:
     sql_clause: sql clause to parse data down
     where_clause: where statement
     sr: spatial reference object
     fields: list of fields to extract from the table
    """
    from arcgis.geometry import _types
    import json
    if HASARCPY:
        sql_clause = kwargs.pop('sql_clause', (None,None))
        where_clause = kwargs.pop('where_clause', None)
        fields = kwargs.pop('fields', None)
        sr = kwargs.pop('sr', None)
        try:
            desc = arcpy.da.Describe(filename)
            area_field = desc.pop('areaFieldName', None)
            length_field = desc.pop('lengthFieldName', None)
        except: # for older versions of arcpy
            desc = arcpy.Describe(filename)
            desc = {
                'fields' : desc.fields,
                'shapeType' : desc.shapeType
            }
            area_field = getattr(desc, 'areaFieldName', None)
            length_field = getattr(desc, 'lengthFieldName', None)
        shape_name = desc['shapeType']
        if fields is None:
            fields = [fld.name for fld in desc['fields'] \
                      if fld.type not in ['Geometry'] and \
                      fld.name not in [area_field, length_field]]
            cursor_fields = fields + ['SHAPE@JSON']
            df_fields = fields + ['SHAPE']
        count = 0
        dfs = []
        shape_field_idx = cursor_fields.index("SHAPE@JSON")
        with da.SearchCursor(filename,
                             field_names=cursor_fields,
                             where_clause=where_clause,
                             sql_clause=sql_clause,
                             spatial_reference=sr) as rows:
            srows = []
            for row in rows:
                srows.append(row)
                if len(srows) == 25000:
                    dfs.append( pd.DataFrame(srows,
                                             columns=df_fields))
                    srows = []
            if len(srows):
                dfs.append( pd.DataFrame(srows,
                                         columns=df_fields))
                srows = []
            del srows
        if len(dfs) > 0:
            df = pd.concat(dfs)
            df = df.reset_index(drop=True)
        elif len(dfs) == 1:
            df = dfs[0]
        else:
            df = pd.DataFrame([],
                              columns=df_fields)
        q = df.SHAPE.notnull()
        gt = desc['shapeType'].lower()
        geoms = {
            "point" : _types.Point,
            "polygon" : _types.Polygon,
            "polyline" : _types.Polyline,
            "multipoint" : _types.MultiPoint,
            "envelope" : _types.Envelope,
            "geometry" : _types.Geometry
        }
        df.SHAPE = (
           df.SHAPE[q]
           .apply(pd.io.json.loads)
           .apply(geoms[gt])
        )
        df.spatial.set_geometry("SHAPE")
        return df
    elif HASARCPY == False and \
         HASPYSHP == True and\
         filename.lower().find('.shp') > -1:
        geoms = []
        records = []
        reader = shapefile.Reader(filename)
        fields = [field[0] for field in reader.fields if field[0] != 'DeletionFlag']
        for r in reader.shapeRecords():
            atr = dict(zip(fields, r.record))
            g = r.shape.__geo_interface__
            g = _geojson_to_esrijson(g)
            geom = _types.Geometry(g)
            atr['SHAPE'] = geom
            records.append(atr)
            del atr
            del r, g
            del geom
        sdf = pd.DataFrame(records)
        sdf.spatial.set_geometry('SHAPE')
        sdf.reset_index(inplace=True)
        return sdf
    elif HASARCPY == False and \
         HASFIONA == True and \
         (filename.lower().find('.shp') > -1 or \
          os.path.dirname(filename).lower().find('.gdb') > -1):
        is_gdb = os.path.dirname(filename).lower().find('.gdb') > -1
        if is_gdb:
            with fiona.drivers():
                from arcgis.geometry import _types
                fp = os.path.dirname(filename)
                fn = os.path.basename(filename)
                geoms = []
                atts = []
                with fiona.open(path=fp, layer=fn) as source:
                    meta = source.meta
                    cols = list(source.schema['properties'].keys())
                    for idx, row in source.items():
                        geoms.append(_types.Geometry(row['geometry']))
                        atts.append(list(row['properties'].values()))
                        del idx, row
                    df = pd.DataFrame(data=atts, columns=cols)
                    df.spatial.set_geometry(geoms)
                    return df
        else:
            with fiona.drivers():
                from arcgis.geometry import _types
                geoms = []
                atts = []
                with fiona.open(path=filename) as source:
                    meta = source.meta
                    cols = list(source.schema['properties'].keys())
                    for idx, row in source.items():
                        geoms.append(_types.Geometry(row['geometry']))
                        atts.append(list(row['properties'].values()))
                        del idx, row
                    df = pd.DataFrame(data=atts, columns=cols)
                    df.spatial.set_geometry(geoms)
                    return df
    return
#--------------------------------------------------------------------------
def to_featureclass(geo,
                    location,
                    overwrite=True,
                    validate=False):
    """
    Exports the DataFrame to a Feature class.

    ===============     ====================================================
    **Argument**        **Description**
    ---------------     ----------------------------------------------------
    location            Required string. This is the output location for the
                        feature class. This should be the path and feature
                        class name.
    ---------------     ----------------------------------------------------
    overwrite           Optional Boolean. If overwrite is true, existing
                        data will be deleted and replaced with the spatial
                        dataframe.
    ===============     ====================================================


    :returns: string

    """


    out_location= os.path.dirname(location)
    fc_name = os.path.basename(location)
    df = geo._data
    if geo.name is None:
        raise ValueError("DataFrame must have geometry set.")
    if validate and \
       geo.validate(strict=True) == False:
        raise ValueError(("Mixed geometry types detected, "
                         "cannot export to feature class."))
    if HASARCPY:
        # 1. Create the Save Feature Class
        #
        columns = df.columns.tolist()
        join_dummy = "AEIOUYAJC81Z"
        columns.pop(columns.index(df.spatial.name))
        dtypes = [(join_dummy, np.int64)]
        if overwrite and arcpy.Exists(location):
            arcpy.Delete_management(location)
        elif overwrite == False and arcpy.Exists(location):
            raise ValueError(('overwrite set to False, Cannot '
                              'overwrite the table. '))

        notnull = geo._data[geo._name].notnull()
        idx = geo._data[geo._name][notnull].first_valid_index()
        sr = sr = geo._data[geo._name][idx]['spatialReference']
        gt = geo._data[geo._name][idx].geometry_type.upper()
        null_geom = {
            'point': pd.io.json.dumps({'x' : None, 'y': None, 'spatialReference' : sr}),
            'polyline' : pd.io.json.dumps({'paths' : [], 'spatialReference' : sr}),
            'polygon' : pd.io.json.dumps({'rings' : [], 'spatialReference' : sr}),
            'multipoint' : pd.io.json.dumps({'points' : [], 'spatialReference' : sr})
        }
        sr = geo._data[geo._name][idx].spatial_reference
        null_geom = null_geom[gt.lower()]
        fc = arcpy.CreateFeatureclass_management(out_location,
                                                 spatial_reference=sr,
                                                 geometry_type=gt,
                                                 out_name=fc_name,
                                                )[0]
        # 2. Add the Fields and Data Types
        #
        oidfld = da.Describe(fc)['OIDFieldName']
        for col in columns[:]:
            if col.lower() in ['fid', 'oid', 'objectid']:
                dtypes.append((col, np.int32))
            elif df[col].dtype.name == 'datetime64[ns]':
                dtypes.append((col, '<M8[us]'))
            elif df[col].dtype.name == 'object':
                try:
                    u = type(df[col][df[col].first_valid_index()])
                except:
                    u = pd.unique(df[col].apply(type)).tolist()[0]
                if issubclass(u, str):
                    mlen = df[col].str.len().max()
                    dtypes.append((col, '<U%s' % int(mlen)))
                else:
                    try:
                        dtypes.append((col, type(df[col][idx])))
                    except:
                        dtypes.append((col, '<U254'))
            elif df[col].dtype.name == 'int64':
                dtypes.append((col, np.int64))
            elif df[col].dtype.name == 'bool':
                dtypes.append((col, np.int32))
            else:
                dtypes.append((col, df[col].dtype.type))

        array = np.array([],
                        np.dtype(dtypes))
        arcpy.da.ExtendTable(fc,
                             oidfld, array,
                             join_dummy, append_only=False)
        # 3. Insert the Data
        #
        fields = arcpy.ListFields(fc)
        icols = [fld.name for fld in fields \
                 if fld.type not in ['OID', 'Geometry'] and \
                 fld.name in df.columns] + ['SHAPE@JSON']
        dfcols = [fld.name for fld in fields \
                  if fld.type not in ['OID', 'Geometry'] and\
                  fld.name in df.columns] + [df.spatial.name]
        with da.InsertCursor(fc, icols) as irows:
            def _insert_row(row):
                row[-1] = pd.io.json.dumps(row[-1])
                irows.insertRow(row)
            q = df[geo._name].isna()
            df.loc[q, 'SHAPE'] = null_geom # set null values to proper JSON
            np.apply_along_axis(_insert_row, 1, df[dfcols].values)
            df.loc[q, 'SHAPE'] = None # reset null values
        return fc
    elif HASPYSHP:
        if fc_name.endswith('.shp') == False:
            fc_name = "%s.shp" % fc_name
        return _pyshp_to_shapefile(df=df,
                            out_path=out_location,
                            out_name=fc_name)
    elif HASARCPY == False and HASPYSHP == False:
        raise Exception(("Cannot Export the data without ArcPy or PyShp modules."
                        " Please install them and try again."))
    else:
        return None
#--------------------------------------------------------------------------
def _pyshp_to_shapefile(df, out_path, out_name):
    """
    Saves a SpatialDataFrame to a Shapefile using pyshp

    :Parameters:
     :df: spatial dataframe
     :out_path: folder location to save the data
     :out_name: name of the shapefile
    :Output:
     path to the shapefile or None if pyshp isn't installed or
     spatial dataframe does not have a geometry column.
    """
    from arcgis.geometry._types import Geometry
    if HASPYSHP:
        GEOMTYPELOOKUP = {
            "Polygon" : shapefile.POLYGON,
            "Point" : shapefile.POINT,
            "Polyline" : shapefile.POLYLINE,
            'null' : shapefile.NULL
        }
        if os.path.isdir(out_path) == False:
            os.makedirs(out_path)
        out_fc = os.path.join(out_path, out_name)
        if out_fc.lower().endswith('.shp') == False:
            out_fc += ".shp"
        geom_field = df.spatial._name
        if geom_field is None:
            return
        geom_type = "null"
        idx = df[geom_field].first_valid_index()
        if idx > -1:
            geom_type = df.loc[idx][geom_field].type
        shpfile = shapefile.Writer(GEOMTYPELOOKUP[geom_type])
        shpfile.autoBalance = 1
        dfields = []
        cfields = []
        for c in df.columns:
            idx = df[c].first_valid_index()
            if idx > -1:
                if isinstance(df[c].loc[idx],
                              Geometry):
                    geom_field = (c, "GEOMETRY")
                else:
                    cfields.append(c)
                    if isinstance(df[c].loc[idx], (str)):
                        shpfile.field(name=c, size=255)
                    elif isinstance(df[c].loc[idx], (int)):
                        shpfile.field(name=c, fieldType="N", size=5)
                    elif isinstance(df[c].loc[idx], (np.int, np.int32, np.int64)):
                        shpfile.field(name=c, fieldType="N", size=10)
                    elif isinstance(df[c].loc[idx], (np.float, np.float64)):
                        shpfile.field(name=c, fieldType="F", size=19, decimal=11)
                    elif isinstance(df[c].loc[idx], (datetime.datetime, np.datetime64)) or \
                         df[c].dtype.name == 'datetime64[ns]':
                        shpfile.field(name=c, fieldType="D", size=8)
                        dfields.append(c)
                    elif isinstance(df[c].loc[idx], (bool, np.bool)):
                        shpfile.field(name=c, fieldType="L", size=1)
            del c
            del idx
        for idx, row in df.iterrows():
            geom = row[df.spatial._name]
            if geom.type == "Polygon":
                shpfile.poly(geom['rings'])
            elif geom.type == "Polyline":
                shpfile.line(geom['paths'])
            elif geom.type == "Point":
                shpfile.point(x=geom.x, y=geom.y)
            else:
                shpfile.null()
            row = row[cfields].tolist()
            for fld in dfields:
                idx = df[cfields].columns.tolist().index(fld)
                if row[idx]:
                    row[idx] = row[idx].to_pydatetime()
            shpfile.record(*row)
            del idx
            del row
            del geom
        shpfile.save(out_fc)


        # create the PRJ file
        try:
            from urllib import request
            wkid = df.spatial.sr['wkid']

            prj_filename = out_fc.replace('.shp', '.prj')

            url = 'http://epsg.io/{}.esriwkt'.format(wkid)

            opener = request.build_opener()
            opener.addheaders = [('User-Agent', 'geosaurus')]
            resp = opener.open(url)

            wkt = resp.read().decode('utf-8')
            if len(wkt) > 0:
                prj = open(prj_filename, "w")
                prj.write(wkt)
                prj.close()
        except:
            # Unable to write PRJ file.
            pass

        del shpfile
        return out_fc
    return None


"""
ArcGIS Server site mode that allows you to control changes to your site.
You can set the site mode to READ_ONLY to disallow the publishing of new
services and block most administrative operations. Your existing services
will continue to function as they did previously. Note that certain
administrative operations such as adding and removing machines from a
site are still available in READ_ONLY mode.
"""
from __future__ import absolute_import
from __future__ import print_function
from .._common import BaseServer

###########################################################################
class Mode(BaseServer):
    """
    ArcGIS Server site mode that allows you to control changes to your site.
    You can set the site mode to READ_ONLY to disallow the publishing of new
    services and block most administrative operations. Your existing services
    will continue to function as they did previously. Note that certain
    administrative operations such as adding and removing machines from a
    site are still available in READ_ONLY mode.
    """
    _url = None
    _con = None
    _json_dict = None
    _json = None
    _siteMode = None
    _copyConfigLocal = None
    _lastModified = None
    #----------------------------------------------------------------------
    def __init__(self,
                 url,
                 gis,
                 initialize=False):
        """Constructor"""
        super(Mode, self).__init__(gis=gis,
                                   url=url)
        if url.lower().endswith('/mode'):
            self._url = url
        else:
            self._url = url + "/mode"
        self._con = gis
        if initialize:
            self._init(gis)
    #----------------------------------------------------------------------
    def update(self,
               siteMode,
               runAsync=False):
        """
        The update operation is used to move between the two types of site
        modes. Switching to READ_ONLY mode will restart all your services
        as the default behavior. Moving to EDITABLE mode will not restart
        services.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        siteMode:           Required string. The mode you will set your site to. Values:
                            READ_ONLY or EDITABLE.
        ---------------     --------------------------------------------------------------------
        runAsync            Optional boolean. Determines if this operation must run asynchronously.
        ===============     ====================================================================


        :return: boolean

        """
        params = {"siteMode" : siteMode,
                  "runAsync" : runAsync,
                  "f" : "json"}
        url = self._url + "/update"
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res




########################################################################
class CategoryManager(object):
    """
    This class allows for the addition, removal and viewing of category
    schema.

    """
    _url = None
    _gis = None
    _con = None

    #----------------------------------------------------------------------
    def __init__(self, gis):
        """Constructor"""
        self._gis = gis
        self._con = gis._con
        baseurl = gis._portal.resturl
        portal_id = None
        if portal_id is None:
            res = self._con.get("%s/portals/self" % baseurl,
                                params={'f': 'json'})
            if 'id' in res:
                pid = res['id']
            else:
                raise Exception("Could not find the portal's ID")
        self._url = "%sportals/%s" % (baseurl, pid)
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    @property
    def schema(self):
        """
        Get/Sets the catagory schema for a GIS.

        When schema is used as a getter, then operation returns the GIS'
        defined category schema is any.

        When schema is used as a setter, the parameter:

        =======================    =============================================================
        **Argument**               **Description**
        -----------------------    -------------------------------------------------------------
        value                      optional list. The schema list.
                                   Syntax Example:
                                   [
                                    {
                                      "title": "Themes",
                                      "categories": [
                                        {
                                          "title": "Basemaps",
                                          "categories": [
                                            {"title": "Partner Basemap"},
                                            {
                                              "title": "Esri Basemaps",
                                              "categories": [
                                                {"title": "Esri Redlands Basemap"},
                                                {"title": "Esri Highland Basemap"}
                                              ]
                                            }
                                          ]
                                        },
                                    {
                                      "title": "Region",
                                      "categories": [
                                        {"title": "US"},
                                        {"title": "World"}
                                      ]
                                    }]}]
        =======================    =============================================================

        """
        url = "%s/categorySchema" % self._url
        params = {'f' : 'json'}
        return self._con.get(url, params)
    #----------------------------------------------------------------------
    @schema.setter
    def schema(self, value):
        """
        Get/Sets the catagory schema for a GIS.

        When schema is used as a getter, then operation returns the GIS'
        defined category schema is any.

        When schema is used as a setter, the parameter:

        =======================    =============================================================
        **Argument**               **Description**
        -----------------------    -------------------------------------------------------------
        value                      optional list. The schema list.
                                   Syntax Example:
                                   [
                                    {
                                      "title": "Themes",
                                      "categories": [
                                        {
                                          "title": "Basemaps",
                                          "categories": [
                                            {"title": "Partner Basemap"},
                                            {
                                              "title": "Esri Basemaps",
                                              "categories": [
                                                {"title": "Esri Redlands Basemap"},
                                                {"title": "Esri Highland Basemap"}
                                              ]
                                            }
                                          ]
                                        },
                                    {
                                      "title": "Region",
                                      "categories": [
                                        {"title": "US"},
                                        {"title": "World"}
                                      ]
                                    }]}]
        =======================    =============================================================

        """
        params = {'f' : 'json'}
        if value is not None:
            params['categorySchema'] = {"categorySchema": value}
            url = "%s/assignCategorySchema" % self._url

            self._con.post(path=url, postdata=params)
        elif value is None:
            url = "%s/deleteCategorySchema" % self._url
            self._con.post(path=url, postdata=params)
    #----------------------------------------------------------------------
    def categorize_item(self, item, categories):
        """
        Assigns or removes a category to a single item.

        =======================    =============================================================
        **Argument**               **Description**
        -----------------------    -------------------------------------------------------------
        item                       Required Item or Item ID (string). The content within a GIS
                                   that will be updated with a list of categories.
        -----------------------    -------------------------------------------------------------
        categories                 Required list. Assigns a list of string values to the item's
                                   categories
        =======================    =============================================================

        :returns: boolean

        """
        from arcgis.gis import Item
        res = []
        if isinstance(item, Item):
            if categories is None:
                res.append(item.update(item_properties={'categories' : ""}))
            else:
                if isinstance(categories, str):
                    categories = [categories]
                res.append(item.update(item_properties={'categories' : categories}))
        return all(res)
    #----------------------------------------------------------------------
    def add(self, items, category):
        """
        Adds a category to an existing set of items

        =======================    =============================================================
        **Argument**               **Description**
        -----------------------    -------------------------------------------------------------
        items                      Required Items. The content within a GIS that will be
                                   updated with a list of categories.
        -----------------------    -------------------------------------------------------------
        category                   Required String. Assigns a category value to the items.
        =======================    =============================================================

        .. code-block:: python

            >>> item = [gis.content.get("<item id 1>"),
                        gis.content.get("<item id 2>")]
            >>> cs = gis.admin.category_schema
            >>> print(cs.add(items=[item], category="/Categories/TEST3"))
            [{'results': [{'itemId': '<item id 1>', 'success': True}]},
             {'results': [{'itemId': '<item id 2>', 'success': True}]}]

        :returns: dict

        """
        from arcgis.gis import Item
        path = self._gis._portal.resturl + "content/updateItems"
        params = {'f' : 'json'}
        updates = []
        results = []
        content = self._gis.content
        if isinstance(items, list):
            for item in items:
                if isinstance(item, Item):
                    categories = item.categories
                    categories.append(category)
                    updates.append({
                        item.itemid : {'categories' : categories}
                    })
                elif isinstance(item, str):
                    item = content.get(item)
                    categories = item.categories
                    categories.append(category)
                    updates.append({
                        item.itemid : {'categories' : categories}
                    })
                del item
        elif isinstance(items, str):
            item = content.get(items)
            categories = item.categories
            categories.append(category)
            updates.append({
                item.itemid : {'categories' : categories}
            })
        elif isinstance(items, Item):
            categories = items.categories
            categories.append(category)
            updates.append({
                items.itemid : {'categories' : categories}
            })
        else:
            raise ValueError("Invalid items, must be list of Item, Item, or item id")
        def _chunks(l, n):
            for i in range(0, len(l), n):
                yield l[i:i+n]
        for i in _chunks(l=updates, n=100):
            params['items'] = i

            res = self._gis._con.post(path=path, postdata=params)
            results.append(res)
            del i
        return results
    #----------------------------------------------------------------------
    def remove(self, items, category):
        """remove a category to an item or items"""
        from arcgis.gis import Item
        path = self._gis._portal.resturl + "content/updateItems"
        params = {'f' : 'json'}
        updates = []
        results = []
        content = self._gis.content
        if isinstance(items, list):
            for item in items:
                if isinstance(item, Item):
                    categories = item.categories
                    if category in categories:
                        del categories[categories.index(category)]
                        updates.append({
                            item.itemid : {'categories' : categories}
                        })
                elif isinstance(item, str):
                    item = content.get(item)
                    categories = item.categories
                    if category in categories:
                        del categories[categories.index(category)]
                        updates.append({
                            item.itemid : {'categories' : categories}
                        })
                del item
        elif isinstance(items, str):
            item = content.get(items)
            categories = item.categories
            if category in categories:
                del categories[categories.index(category)]
                updates.append({
                    item.itemid : {'categories' : categories}
                })
        elif isinstance(items, Item):
            categories = items.categories
            if category in categories:
                del categories[categories.index(category)]
                updates.append({
                    items.itemid : {'categories' : categories}
                })
        else:
            raise ValueError("Invalid items, must be list of Item, Item, or item id")
        def _chunks(l, n):
            for i in range(0, len(l), n):
                yield l[i:i+n]
        for i in _chunks(l=updates, n=100):
            params['items'] = i

            res = self._gis._con.post(path=path, postdata=params)
            results.append(res)
            del i
        return results
    #----------------------------------------------------------------------
    def replace(self, items, old_category, new_catgory):
        """finds and replaces a category value with a new value one"""
        res = self.add(items, new_catgory)
        res = self.remove(items, old_category)
        return res

    #----------------------------------------------------------------------
    def reset(self, items):
        """deletes all the categories for a given set of items"""
        return self._gis.content.bulk_update(items, {'categories': ''})




__version__ = '1.5.2'

from . import features, geoanalytics, geocoding, geometry, geoprocessing, network, raster, realtime, schematics, mapping, apps

from .gis import GIS
from .features.analysis import *
from .geocoding import geocode

try:
    import pandas as pd
    from .features.geo import GeoAccessor, GeoSeriesAccessor
except:
    pass

__all__ = ['GIS', 'geocode', 'features',
           'geoanalytics', 'geocoding', 'geometry',
           'geoprocessing', 'network', 'raster', 'apps',
           'realtime', 'schematics', 'mapping',
           'aggregate_points',
           'calculate_density',
           'connect_origins_to_destinations',
           'create_buffers',
           'create_drive_time_areas',
           'create_route_layers',
           'create_viewshed',
           'create_watersheds',
           'derive_new_locations',
           'dissolve_boundaries',
           'enrich_layer',
           'extract_data',
           'find_existing_locations',
           'find_hot_spots',
           'find_nearest',
           'find_similar_locations',
           'interpolate_points',
           'join_features',
           'merge_layers',
           'overlay_layers',
           'plan_routes',
           'summarize_nearby',
           'summarize_within',
           'trace_downstream']

def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'widgets/js/dist',
        'dest': 'arcgis',
        'require': 'arcgis/extension'
    }]




"""
Defines the AssignmentType class.
"""

from .exceptions import ValidationError
from .model import Model
from ._store import *


class AssignmentType(Model):
    """
    Defines the acceptable values for :class:`~arcgis.apps.workforce.Assignment` types.

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    project                Required :class:`~arcgis.apps.workforce.Project`. The project that
                           this assignment belongs to.
    ------------------     --------------------------------------------------------------------
    coded_value            Optional :class:`dict`. The dictionary storing the code and
                           name of the type.
    ------------------     --------------------------------------------------------------------
    name                   Optional :class:`String`. The name of the assignment type.
    ==================     ====================================================================

    """

    def __init__(self, project, coded_value=None, name=None):
        super().__init__()
        self.project = project
        if coded_value:
            self._coded_value = coded_value
        else:
            self._coded_value = {'code': None, 'name': name}

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<AssignmentType {}>".format(self.code)

    def update(self, name=None):
        """
            Updates the assignment type on the server

            ==================     ====================================================================
            **Argument**           **Description**
            ------------------     --------------------------------------------------------------------
            name                   Optional :class:`String`.
                                   The name of the assignment type
            ==================     ====================================================================
        """
        update_assignment_type(self.project, self, name)

    def delete(self):
        """Deletes the assignment type from the server"""
        delete_assignment_types(self.project, [self])

    @property
    def id(self):
        """Gets the id of the assignment type"""
        return self.code

    @property
    def code(self):
        """Gets the internal code that uniquely identifies the assignment type"""
        return self._coded_value['code']

    @property
    def name(self):
        """Gets/Sets The name of the assignment type"""
        return self._coded_value['name']

    @name.setter
    def name(self, value):
        self._coded_value['name'] = value

    @property
    def coded_value(self):
        """Gets the coded value"""
        return self._coded_value

    def _validate(self, **kwargs):
        errors = super()._validate(**kwargs)
        errors += self._validate_name()
        errors += self._validate_name_uniqueness(**kwargs)
        return errors

    def _validate_for_update(self, **kwargs):
        return super()._validate_for_update(**kwargs) + self._validate_code()

    def _validate_for_remove(self, **kwargs):
        assignments = kwargs['assignments']
        errors = super()._validate_for_remove(**kwargs) + self._validate_code()
        if assignments is None:
            schema = self.project._assignment_schema
            where = "{}={}".format(schema.assignment_type, self.code)
            assignments = self.project.assignments.search(where=where)
        else:
            assignments = [a for a in assignments if a.assignment_type.code == self.code]
        if assignments:
            errors.append(ValidationError("Cannot remove an in-use AssignmentType", self))
        return errors

    def _validate_name(self):
        errors = []
        if self.name is None or self.name.isspace():
            errors.append(ValidationError("AssignmentType must have a name", self))
        elif '>' in self.name or '<' in self.name or '%' in self.name:
            errors.append(ValidationError("AssignmentType name contains invalid characters", self))
        return errors

    def _validate_name_uniqueness(self, assignment_types=None):
        errors = []
        if assignment_types is None:
            assignment_types = self.project.assignment_types.search()
        for assignment_type in assignment_types:
            if assignment_type.name == self.name and assignment_type.code != self.code:
                errors.append(ValidationError("AssignmentType name must be unique", self))
        return errors

    def _validate_code(self):
        errors = []
        if not isinstance(self.code, int):
            errors.append(ValidationError("Code must be a unique integer", self))
        return errors



"""
The arcgis.network module contains classes and functions for network analysis. Network layers and analysis can be used
for operations such as finding the closest facility, the best route for a vehicle, the best routes for a fleet of
vehicles, locating facilities using location allocation, calculating an OD cost matrix, and generating service areas.
"""

from ._layer import NetworkLayer, NetworkDataset, ClosestFacilityLayer, ServiceAreaLayer, RouteLayer

from . import analysis



reference = {
    'journal' : {
        "values":{
            "settings":{"appGeocoders":[{"singleLineFieldName":"SingleLine",
                                         "name":"Esri World Geocoder",
                                         "url":"https://geocode.arcgis.com/arcgis/rest/services/World/GeocodeServer"},
                                        {"singleLineFieldName":"SingleLine",
                                         "name":"san",
                                         "placeholder":"",
                                         "url":"https://sampleserver5.arcgisonline.com/arcgis/rest/services/Locators/SanDiego/GeocodeServer"}
                                        ],
                        "layout":{"id":"side"}},
            "title": "title",
            "story" : {
                "storage": "WEBAPP",
                "sections": []
                },
            "template":{"name":"Map Journal",
                        "createdWith":"1.14.1",
                        "editedWith":"1.14.1"}
        }
    }
}



"""
The **env** module provides a shared environment used by the different modules.
It stores globals such as the currently active GIS, the default geocoder and so on.
It also stores environment settings that are common among all geoprocessing tools,
such as the output spatial reference.

active_gis
==========

.. py:data:: active_gis
The currently active GIS, that is used for analysis functions unless explicitly specified
when calling the functions.
Creating a new GIS object makes it active unless set_active=False is passed in the GIS constructor.


analysis_extent
===============

.. py:data:: analysis_extent
The processing extent used by analysis tools, specified as an arcgis.geometry.Envelope.

out_spatial_reference
=====================

.. py:data:: out_spatial_reference
The spatial reference of the output geometries. If not specified, the output geometries are in the
spatial reference of the input geometries. If process_spatial_reference is specified and out_spatial_reference
is not specified, the output geometries are in the spatial reference of the process spatial reference.

process_spatial_reference
=========================

.. py:data:: process_spatial_reference
The spatial reference that the geoprocessor will use to perform geometry operations. If specified and
out_spatial_reference is not specified, the output geometries are in the spatial reference of the
process spatial reference.

output_datastore
================

.. py:data:: output_datastore
The data store where GeoAnalytics results should be stored. The supported values of this parameter are "relational" and
"spatiotemporal". By default, results are stored in the spatiotemporal data store. It is recommended that results are
stored in the spatiotemporal data store due to the scalability of the spatiotemporal big data store.

return_z
========

.. py:data:: return_z
If true, Z values will be included in the geoprocessing results if the features have Z values.
Otherwise Z values are not returned. The default is False.


return_m
========

.. py:data:: return_m
If true, M values will be included in the results if the features have M values.
Otherwise M values are not returned. The default is False.

verbose
========

.. py:data:: verbose

If True, messages from geoprocessing tools will be printed to stdout.
In any case, all geoprocessing messages are available through Python logging module.

default_aggregation_styles
==========================

.. py:data:: default_aggregation_styles

Tasks that have the default_aggregation_styles property set to true will set the default aggregations for the
resulting layer. Default aggregations can be square, pointy triangle, flat triangle, pointy hexagon, flat
hexagon, and geohash. All aggregation styles are supported using WKID 4326 (WGS_1984).
The default_aggregation_styles is False. (supported at 10.6.1+)

snap_raster
===========

.. py:data:: snap_raster

Tasks that honor the snap_raster environment will adjust the extent of output rasters 
so that they match the cell alignment of the specified snap raster.
(For more information about this environment setting, 
please refer to http://pro.arcgis.com/en/pro-app/tool-reference/environment-settings/snap-raster.htm)

cell_size
=========

.. py:data:: cell_size

Tasks that honor the Cell size environment setting set the output raster cell size, or resolution, 
for the operation.  The default output resolution is determined by the largest cell size of 
all the input rasters.
(For more information about this environment setting, 
please refer to http://pro.arcgis.com/en/pro-app/tool-reference/environment-settings/cell-size.htm)

mask
====

.. py:data:: mask

Tasks that honor the mask environment will only consider those cells that fall within the analysis 
mask in the operation
(For more information about this environment setting, 
please refer to http://pro.arcgis.com/en/pro-app/tool-reference/environment-settings/mask.htm)

parallel_processing_factor
==========================

.. py:data:: parallel_processing_factor

Tasks that honor the parallel_processing_factor environment will divide and perform operations across 
multiple processes.
(For more information about this environment setting, 
please refer to http://pro.arcgis.com/en/pro-app/tool-reference/environment-settings/parallel-processing-factor.htm)
"""

#: The currently active GIS, that is used for analysis functions unless explicitly specified.
#: Creating a new GIS object makes it active by default unless set_active=False is passed in the GIS constructor.
active_gis = None

#: The spatial reference of the output geometries. If not specified, the output geometries are in the
#: spatial reference of the input geometries. If process_spatial_reference is specified and out_spatial_reference
#: is not specified, the output geometries are in the spatial reference of the process spatial reference.
out_spatial_reference = None

#: The spatial reference that analysis and geoprocessing tools will use to perform geometry operations. If specified and
#: out_spatial_reference is not specified, the output geometries are in the spatial reference of the
#: process spatial reference.
process_spatial_reference = None

#: The data store where GeoAnalytics results should be stored. The supported values of this parameter are "relational" and
#: "spatiotemporal". By default, results are stored in the spatiotemporal data store. It is recommended that results be
#: stored in the spatiotemporal data store due to the scalability of the spatiotemporal big data store.
output_datastore = None

#: The processing extent used by analysis tools
analysis_extent = None

#: If True, Z values will be included in the geoprocessing results if the features have Z values.
#: Otherwise Z values are not returned. The default is False.
return_z = False

#: If True, M values will be included in the results if the features have M values.
#: Otherwise M values are not returned. The default is False.
return_m = False

#: If True, messages from geoprocessing tools will be printed to stdout
verbose = False

#: Tasks that have the default_aggregation_styles property set to true will set the default aggregations for the
#: resulting layer. Default aggregations can be square, pointy triangle, flat triangle, pointy hexagon, flat
#: hexagon, and geohash. All aggregation styles are supported using WKID 4326 (WGS_1984).
#: The default_aggregation_styles is False. (supported at 10.6.1+)
default_aggregation_styles = False

#: Tasks that honor the snap_raster environment will adjust the extent of output rasters 
#: so that they match the cell alignment of the specified snap raster.
snap_raster = None

#: Tasks that honor the cell_size environment setting set the output raster cell size, or resolution, for the operation.
#: The default output resolution is determined by the largest cell size of all the input rasters
cell_size = None

#: Tasks that honor the mask environment will only consider those cells that fall within the analysis mask in the operation
mask = None

#Tasks that honor the parallel_processing_factor environment will divide and perform operations across multiple processes.
parallel_processing_factor = None





from arcgis.features import Feature, FeatureSet
from arcgis.features import FeatureLayer, Table
from arcgis.geometry import Geometry
import pandas as pd
# --------------------------------------------------------------------------
def _chunks(l, n):
    """yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        yield l[i:i + n]
# --------------------------------------------------------------------------
_look_up_types = {
    "esriFieldTypeBlob" : "object",
    "esriFieldTypeDate" : "datetime64",
    "esriFieldTypeInteger" : "int64",
    "esriFieldTypeSmallInteger" : "int32",
    "esriFieldTypeDouble" : "float64",
    "esriFieldTypeSingle" :  "float32",
    "esriFieldTypeString" : "str",
    "esriFieldTypeGeometry" : "object",
    "esriFieldTypeOID" : "int64",
    "esriFieldTypeGlobalID" : "str",
    "esriFieldTypeRaster" : "object",
    "esriFieldTypeGUID" : "str",
    "esriFieldTypeXML" : "object"
}
# --------------------------------------------------------------------------
def to_featureset(df):
    """converts a pd.DataFrame to a FeatureSet Object"""
    if hasattr(df, 'spatial'):
        fs = df.spatial.__feature_set__
        return FeatureSet.from_dict(fs)
    return None
# --------------------------------------------------------------------------
def from_featureset(fset, sr=None):
    """

    Converts a FeatureSet to a pd.DataFrame

    ===============    ==============================================
    Arguments          Description
    ---------------    ----------------------------------------------
    fset               Required FeatureSet.  FeatureSet object.
    ===============    ==============================================

    return Panda's DataFrame

    """
    if isinstance(fset, FeatureSet):
        rows = []
        sr = fset.spatial_reference
        dt_fields = [fld['name'] for fld in fset.fields if fld['type'] == 'esriFieldTypeDate']
        if sr is None:
            sr = {'wkid':4326}
        for feat in fset.features:
            a = feat.attributes
            if not feat.geometry is None:
                g = feat.geometry
                g['spatialReference'] = sr
                a['SHAPE'] = Geometry(g)
            rows.append(a)
            del a, feat
        from arcgis.features import GeoAccessor, GeoSeriesAccessor
        df = pd.DataFrame(data=rows)
        for fld in dt_fields:
            try:
                df[fld] = pd.to_datetime(df[fld]/1000,
                                         infer_datetime_format=True,
                                         unit='s')
            except:
                df[fld] = pd.to_datetime(df[fld], infer_datetime_format=True)
        if 'SHAPE' in df.columns:
            df.spatial.set_geometry("SHAPE")
            df.spatial.sr = sr
        return df
    else:
        return None
#--------------------------------------------------------------------------
def from_layer(layer,
               query="1=1"):
    """
    Converts a Feature Service Layer to a Pandas' DataFrame

    Parameters:
     :layer: FeatureLayer or Table object.  If the object is a FeatureLayer
      the function will return a Spatial DataFrame, if the object is of
      type Table, the function will return a Pandas' DataFrame


    Usage:

    >>> from arcgis.features import FeatureLayer
    >>> mylayer = FeatureLayer(("https://sampleserver6.arcgisonline.com/arcgis/rest"
                        "/services/CommercialDamageAssessment/FeatureServer/0"))
    >>> df = from_layer(mylayer)
    >>> print(df.head())

    """
    if not layer.filter is None:
        query = layer.filter
    from arcgis.geometry import Geometry, SpatialReference
    fields = []
    records = []
    if isinstance(layer, (Table, FeatureLayer)) == False:
        raise ValueError("Invalid inputs: must be FeatureLayer or Table")
    if 'maxRecordCount' in layer.properties:
        max_records = layer.properties['maxRecordCount']
    else:
        max_records = 1000
    service_count = layer.query(where=query,
                                return_count_only=True)
    if service_count > max_records:
        frames = []
        oid_info = layer.query(where=query, return_ids_only=True)
        for ids in _chunks(oid_info['objectIds'], max_records):
            ids = [str(i) for i in ids]
            sql = "%s in (%s)" % (oid_info['objectIdFieldName'],
                                  ",".join(ids))
            frames.append(layer.query(where=sql, as_df=True))
        res = pd.concat(frames, ignore_index=True)
        res.reset_index(drop=True, inplace=True)
        res.spatial.set_geometry("SHAPE")
    else:
        return layer.query(where=query, as_df=True)#.sdf
    return res
#----------------------------------------------------------------------
def to_layer(df,
             layer,
             update_existing=True,
             add_new=False,
             truncate=False):
    """
    Sends the Spatial DataFrame information to a published service

    :Parameters:
     :df: Spatial DataFrame object
     :layer: Feature Layer or Table Layer object
     :update_existing: boolean -
     :add_new: boolean
     :truncate: if true, all records will be deleted and the dataframe
      records will replace the service data
    Output:
     A layer object
    """
    if not isinstance(df, (pd.DataFrame)) or not hasattr(df, 'spatial'):
        raise ValueError("df must be a SpatialDatframe")
    if not isinstance(layer, (Table, FeatureLayer)):
        raise ValueError("layer must be a FeatureLayer or Table Layer")
    if truncate:
        layer.delete_features(where='1=1')
        layer.edit_features(adds=to_featureset(df).features)
    elif update_existing:
        layer.edit_features(updates=to_featureset(df).features)
    elif add_new:
        layer.edit_features(adds=to_featureset(df).features)
    return layer



import json

########################################################################
class Extension(object):
    """
    represents a service extension
    """
    _typeName = None
    _capabilities = None
    _enabled = None
    _maxUploadFileSize = None
    _allowedUploadFileTypes = None
    _properties = None
    _allowedExtensions = ["naserver", "mobileserver",
                          "kmlserver", "wfsserver",
                          "schematicsserver", "featureserver",
                          "wcsserver", "wmsserver"]
    #----------------------------------------------------------------------
    def __init__(self, type_name,
                 capabilities,
                 enabled,
                 max_upload_file_size,
                 allowed_upload_filetype,
                 properties):
        """Constructor"""
        self._typeName = type_name
        self._capabilities = capabilities
        self._enabled = enabled
        self._maxUploadFileSize = max_upload_file_size
        self._allowedUploadFileTypes = allowed_upload_filetype
        self._properties = properties
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """gets/sets the extension properties"""
        return self._properties
    #----------------------------------------------------------------------
    @properties.setter
    def properties(self, value):
        """gets/sets the extension properties"""
        if isinstance(value, dict):
            self._properties = value
    #----------------------------------------------------------------------
    @property
    def typeName(self):
        """gets the extension type"""
        return self._typeName
    #----------------------------------------------------------------------
    @property
    def capabilities(self):
        """gets/sets the capabilities"""
        return self._capabilities
    #----------------------------------------------------------------------
    @capabilities.setter
    def capabilities(self, value):
        """gets/sets the capabilities"""
        if self._capabilities != value:
            self._capabilities = value
    #----------------------------------------------------------------------
    @property
    def enabled(self):
        """gets/sets the extension is enabled"""
        return self._enabled
    #----------------------------------------------------------------------
    @enabled.setter
    def enabled(self, value):
        """gets/sets the extension is enabled"""
        if isinstance(value, bool):
            self._enabled = value
    #----------------------------------------------------------------------
    @property
    def max_upload_file_size(self):
        """sets/gets the maxUploadFileSize"""
        return self._maxUploadFileSize
    #----------------------------------------------------------------------
    @max_upload_file_size.setter
    def max_upload_file_size(self, value):
        """sets/gets the maxUploadFileSize"""
        if isinstance(value, int):
            self._maxUploadFileSize = value
    #----------------------------------------------------------------------
    @property
    def allowed_upload_filetypes(self):
        """gets/sets the allowedUploadFileTypes"""
        return self._allowedUploadFileTypes
    #----------------------------------------------------------------------
    @allowed_upload_filetypes.setter
    def allowed_upload_filetypes(self, value):
        """gets/sets the allowedUploadFileTypes"""
        self._allowedUploadFileTypes = value
    #----------------------------------------------------------------------
    def __str__(self):
        """returns the object as JSON"""
        return json.dumps({
            "typeName": self._typeName,
            "capabilities": self._capabilities,
            "enabled": self._enabled,
            "maxUploadFileSize": self._maxUploadFileSize,
            "allowedUploadFileTypes": self._allowedUploadFileTypes,
            "properties": self._properties
        })
    #----------------------------------------------------------------------
    @property
    def value(self):
        """returns the object as a dictionary"""
        return json.loads(str(self))

    #----------------------------------------------------------------------
    @staticmethod
    def fromJSON(value):
        """returns the object from json string or dictionary"""
        if isinstance(value, str):
            value = json.loads(value)
        elif isinstance(value, dict):
            value = value
        else:
            raise AttributeError("Invalid input")
        if 'allowedUploadFileTypes' not in value:
            value['allowedUploadFileTypes'] = ""
        return Extension(type_name=value['typeName'],
                         capabilities=value['capabilities'] or "",
                         enabled=value['enabled'] == "true",
                         max_upload_file_size=value['maxUploadFileSize'],
                         allowed_upload_filetype=value['allowedUploadFileTypes'] or "",
                         properties=value['properties'])
########################################################################
class ClusterProtocol(object):
    """
    The clustering protocol defines a channel which is used by server
    machines within a cluster to communicate with each other. A server
    machine will communicate with its peers information about the status of
    objects running within it for load balancing and fault tolerance.
    ArcGIS Server supports the TCP clustering protocols where server
    machines communicate with each other over a TCP channel (port).

    Inputs:
       tcpClusterPort - The port to use when configuring a TCP based
       protocol. By default, the server will pick up the next value in the
       assigned ports on all machines.
    """
    _tcpClusterPort = None
    #----------------------------------------------------------------------
    def __init__(self, tcpClusterPort):
        """Constructor"""
        self._tcpClusterPort = int(tcpClusterPort)
    #----------------------------------------------------------------------
    @property
    def tcpClusterPort(self):
        """
        The port to use when configuring a TCP based protocol. By default,
        the server will pick up the next value in the assigned ports on
        all machines.
        """
        return self._tcpClusterPort
    #----------------------------------------------------------------------
    def __str__(self):
        """"""
        return json.dumps({
            "tcpClusterPort" : self._tcpClusterPort
        })
    #----------------------------------------------------------------------
    @property
    def value(self):
        """
        returns the tcpClusterPort as a dictionary
        """
        return {
            "tcpClusterPort" : self._tcpClusterPort
        }









import logging as _logging
import arcgis
from datetime import datetime
from arcgis.features import FeatureSet
from arcgis.mapping import MapImageLayer
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData
from arcgis.geoprocessing._support import _execute_gp_tool

_log = _logging.getLogger(__name__)

_use_async = False


def get_travel_modes(
    gis = None):
    """


Get a list of travel modes that can be used with directions and routing services available in your portal.

Parameters:

gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns the following as a named tuple:
   supported_travel_modes - Supported Travel Modes as a FeatureSet
   default_travel_mode - Default Travel Mode as a str

See https://logistics.arcgis.com/arcgis/rest/directories/arcgisoutput/World/Utilities_GPServer/World_Utilities/GetTravelModes.htm for additional help.
    """
    kwargs = locals()

    param_db = {
        "supported_travel_modes": (FeatureSet, "Supported Travel Modes"),
        "default_travel_mode": (str, "Default Travel Mode"),
    }
    return_values = [
        {"name": "supported_travel_modes", "display_name": "Supported Travel Modes", "type": FeatureSet},
        {"name": "default_travel_mode", "display_name": "Default Travel Mode", "type": str},
    ]

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.routingUtilities.url

    return _execute_gp_tool(gis, "GetTravelModes", kwargs, param_db, return_values, _use_async, url)

get_travel_modes.__annotations__ = {
    'return': tuple}
    
def get_tool_info(
    service_name = """asyncRoute""",
    tool_name = """FindRoutes""",
    gis = None):
    """


Get additional information such as the description of the network dataset used for the analysis and the execution limits for a tool in a geoprocessing service.

Parameters:

   service_name: Service Name (str). Required parameter.  Specify the service name containing the tool. The parameter value should be specified using one of the following keywords that reference a particular geoprocessing service.asyncClosestFacility - The asynchronous geoprocessing service used to perform the closest facility analysis.asyncLocationAllocation - The asynchronous geoprocessing service used to perform the location-allocation analysis.asyncRoute - The asynchronous geoprocessing service used to perform the route analysis.asyncServiceArea - The asynchronous geoprocessing service used to perform the service area analysis.asyncVRP - The asynchronous geoprocessing service used to perform the vehicle routing problem analysis.syncVRP - The synchronous geoprocessing service used to perform the vehicle routing problem analysis.The default value is asyncRoute.
      Choice list:['asyncClosestFacility', 'asyncLocationAllocation', 'asyncODCostMatrix', 'asyncRoute', 'asyncServiceArea', 'asyncVRP', 'syncVRP']

   tool_name: Tool Name (str). Required parameter.  Specify the tool name in the geoprocessing service. The parameter value should be a valid tool name in the geoprocessing service specified by the serviceName parameter. The default value is FindRoutes.
      Choice list:['EditVehicleRoutingProblem', 'FindClosestFacilities', 'FindRoutes', 'GenerateOriginDestinationCostMatrix', 'GenerateServiceAreas', 'SolveLocationAllocation', 'SolveVehicleRoutingProblem']

gis: Optional, the GIS on which this tool runs. If not specified, the active GIS is used.


Returns:
   tool_info - Tool Info as a str

See https://logistics.arcgis.com/arcgis/rest/directories/arcgisoutput/World/Utilities_GPServer/World_Utilities/GetToolInfo.htm for additional help.
    """
    kwargs = locals()

    param_db = {
        "service_name": (str, "serviceName"),
        "tool_name": (str, "toolName"),
        "tool_info": (str, "Tool Info"),
    }
    return_values = [
        {"name": "tool_info", "display_name": "Tool Info", "type": str},
    ]

    if gis is None:
        gis = arcgis.env.active_gis

    url = gis.properties.helperServices.routingUtilities.url
    return _execute_gp_tool(gis, "GetToolInfo", kwargs, param_db, return_values, _use_async, url)

get_tool_info.__annotations__ = {
    'service_name': str,
    'tool_name': str,
    'return': str}    


"""
This resource is a collection of all the items that have been uploaded
to the server.

There are two ways to upload items. You can upload complete items using
the Upload Item operation. If a particular item is made up of many
chunks (parts), you need to first register the item and subsequently
upload the individual parts using the Upload Part operation. Item
uploads are filtered by a whitelist of filename extensions. This is the
default list: soe, sd, sde, odc, csv, txt, zshp, kmz. The default list
can be overridden by setting the uploadFileExtensionWhitelist property
with the server properties API.

"""
from __future__ import absolute_import
from __future__ import print_function
from .._common import BaseServer
from six.moves.urllib_parse import urlparse
########################################################################
class Uploads(BaseServer):
    """
    This resource is a collection of all the items that have been uploaded
    to the server.

    There are two ways to upload items. You can upload complete items using
    the Upload Item operation. If a particular item is made up of many
    chunks (parts), you need to first register the item and subsequently
    upload the individual parts using the Upload Part operation. Item
    uploads are filtered by a whitelist of filename extensions. This is the
    default list: soe, sd, sde, odc, csv, txt, zshp, kmz. The default list
    can be overridden by setting the uploadFileExtensionWhitelist property
    with the server properties API.

    """
    _uploads = None
    _con = None
    _json = None
    _json_dict = None
    _url = None

    #----------------------------------------------------------------------
    def __init__(self, url, gis,
                 initialize=False):
        """Constructor"""
        if url.lower().find("uploads") < -1:
            self._url = url + "/uploads"
        else:
            self._url = url
        self._con = gis
        self._json_dict = {}
        self._json = ""

    #----------------------------------------------------------------------
    @property
    def uploads(self):
        """
        returns a collection of all the items that have been uploaded to
        the server.

        There are two ways to upload items. You can upload complete items
        using the Upload Item operation. If a particular item is made up of
        many chunks (parts), you need to first register the item and
        subsequently upload the individual parts using the Upload Part
        operation. Item uploads are filtered by a whitelist of filename
        extensions. This is the default list: soe, sd, sde, odc, csv, txt,
        zshp, kmz. The default list can be overridden by setting the
        uploadFileExtensionWhitelist property with the server properties API.

        """
        params = {
            "f" :"json"
        }
        return self._con.get(path=self._url,
                            params=params)
    #----------------------------------------------------------------------
    def delete(self, item_id):
        """
        Deletes the uploaded item and its configuration.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        item_id             Required string. unique ID of the item
        ===============     ====================================================================


        :return: boolean

        """
        url = self._url + "/%s/delete" % item_id
        params = {
            "f" : "json"
        }
        res = self._con.post(path=url, postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res
    #----------------------------------------------------------------------
    def item(self, item_id):
        """
        This resource represents an item that has been uploaded to the
        server. Various workflows upload items and then process them on the
        server. For example, when publishing a GIS service from ArcGIS for
        Desktop or ArcGIS Server Manager, the application first uploads the
        service definition (.SD) to the server and then invokes the
        publishing geoprocessing tool to publish the service.
        Each uploaded item is identified by a unique name (item_id). The
        pathOnServer property locates the specific item in the ArcGIS
        Server system directory.
        The committed parameter is set to true once the upload of
        individual parts is complete.

        Parameters:
         :item_id: uploaded id identifier
        """
        url = self._url + "/%s" % item_id
        params = {
            "f" : "json"
        }
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    def upload(self, path, description=None):
        """
        Uploads a new item to the server. Once the operation is completed
        successfully, the JSON structure of the uploaded item is returned.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        path                Required string. The file location to upload
        ---------------     --------------------------------------------------------------------
        description         Optional string. Description of the upload.
        ===============     ====================================================================


        :return: boolean


        """
        url = self._url + "/upload"
        params = {
            "f" : "json"
        }
        files = {}
        files['itemFile'] = path
        if description:
            params['description'] = description
        res = self._con.post(path=url,
                             postdata=params,
                             files=files)
        if 'status' in res and \
           res['status'] == 'success':
            return True, res
        return False, res
    #----------------------------------------------------------------------
    def _service_configuration(self, upload_id):
        """ gets the serviceconfiguration.json info for an uploaded sd file"""
        url = self._url + "/%s/serviceconfiguration.json" % upload_id
        params = {'f' : 'json'}
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    def _initial_cache_settings(self, upload_id):
        """ gets the initial cache settings for a given uploaded sd file"""
        url = self._url + "/%s/serviceconfiguration.json" % upload_id
        params = {'f' : 'json'}
        return self._con.get(path=url, params=params)
    #----------------------------------------------------------------------
    def upload_by_part(self,
                       item_id,
                       part_number,
                       part):
        """
        Uploads a new item to the server. Once the operation is completed
        successfully, the JSON structure of the uploaded item is returned.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        item_id             Required string. Item ID to upload to.
        ---------------     --------------------------------------------------------------------
        part_number         Required int. An integer value associated with the part.
        ---------------     --------------------------------------------------------------------
        part                Required string. File path to the part to upload.
        ===============     ====================================================================


        :return: dict

        """
        url = self._url + "{iid}/uploadPart".format(iid=item_id)
        params = {
            "f" : "json",
            "partNumber" : part_number
        }
        files = {}
        files['partFile'] = part
        return self._con.post(path=url,
                              postdata=params,
                              files=files)
    #----------------------------------------------------------------------
    def commit(self, item_id, parts=None):
        """
        Use this operation to complete the upload of all the parts that
        make an item. The parts parameter indicates to the server all the
        parts that make up the item.

        ===============     ====================================================================
        **Argument**        **Description**
        ---------------     --------------------------------------------------------------------
        item_id             Required string. Item ID to commit.
        ---------------     --------------------------------------------------------------------
        parts               Optional list. An optional comma-separated ordered list of all the
                            parts that make the item. If this parameter is not provided, the
                            default order of the parts is used.
        ===============     ====================================================================


        :return: Boolean

        """
        params = {'f': 'json'}
        url = self._url + '/{iid}/commit'.format(iid=item_id)
        if parts:
            params['parts'] = parts
        res = self._con.post(path=url,
                             postdata=params)
        if 'status' in res:
            return res['status'] == 'success'
        return res



"""
Functions which take geometric types as parameters and return geometric type results.
"""

import arcgis.env

# https://utility.arcgisonline.com/ArcGIS/rest/services/Geometry/GeometryServer.

def areas_and_lengths(polygons,
                      length_unit,
                      area_unit,
                      calculation_type,
                      spatial_ref=4326,
                      gis=None):
    """
       The areas_and_lengths function calculates areas and perimeter lengths
       for each polygon specified in the input array.

       Inputs:
          polygons - The array of polygons whose areas and lengths are
                     to be computed.
          length_unit - The length unit in which the perimeters of
                       polygons will be calculated. If calculation_type
                       is planar, then length_unit can be any esriUnits
                       constant. If lengthUnit is not specified, the
                       units are derived from spatial_ref. If calculationType is
                       not planar, then lengthUnit must be a linear
                       esriUnits constant, such as esriSRUnit_Meter or
                       esriSRUnit_SurveyMile. If length_unit is not
                       specified, the units are meters. For a list of
                       valid units, see esriSRUnitType Constants and
                       esriSRUnit2Type Constant.
          area_unit - The area unit in which areas of polygons will be
                     calculated. If calculation_type is planar, then
                     area_unit can be any esriUnits constant. If
                     area_unit is not specified, the units are derived
                     from spatial_ref. If calculation_type is not planar, then
                     area_unit must be a linear esriUnits constant such
                     as esriSRUnit_Meter or esriSRUnit_SurveyMile. If
                     area_unit is not specified, then the units are
                     meters. For a list of valid units, see
                     esriSRUnitType Constants and esriSRUnit2Type
                     constant.
                     The list of valid esriAreaUnits constants include,
                     esriSquareInches | esriSquareFeet |
                     esriSquareYards | esriAcres | esriSquareMiles |
                     esriSquareMillimeters | esriSquareCentimeters |
                     esriSquareDecimeters | esriSquareMeters | esriAres
                     | esriHectares | esriSquareKilometers.
          calculation_type -  The type defined for the area and length
                             calculation of the input geometries. The
                             type can be one of the following values:
                             planar - Planar measurements use 2D
                                      Euclidean distance to calculate
                                      area and length. Th- should
                                      only be used if the area or
                                      length needs to be calculated in
                                      the given spatial reference.
                                      Otherwise, use preserveShape.
                             geodesic - Use this type if you want to
                                      calculate an area or length using
                                      only the vertices of the polygon
                                      and define the lines between the
                                      points as geodesic segments
                                      independent of the actual shape
                                      of the polygon. A geodesic
                                      segment is the shortest path
                                      between two points on an ellipsoid.
                             preserveShape - This type calculates the
                                      area or length of the geometry on
                                      the surface of the Earth
                                      ellipsoid. The shape of the
                                      geometry in its coordinate system
                                      is preserved.
       Output:
          JSON as dictionary
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.areas_and_lengths(
        polygons,
        length_unit,
        area_unit,
        calculation_type,
        spatial_ref)


def auto_complete(polygons=None,
                  polylines=None,
                  spatial_ref=None,
                  gis=None):
    """
       The auto_complete function simplifies the process of
       constructing new polygons that are adjacent to other polygons.
       It constructs polygons that fill in the gaps between existing
       polygons and a set of polylines.

       Inputs:
          polygons - array of Polygon objects
          polylines - list of Polyline objects
          spatial_ref - spatial reference of the input geometries WKID
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.auto_complete(
        polygons,
        polylines,
        spatial_ref)

def buffer(geometries,
           in_sr,
           distances,
           unit,
           out_sr=None,
           buffer_sr=None,
           union_results=None,
           geodesic=None,
           gis=None):
    """
       The buffer function is performed on a geometry service resource
       The result of this function is buffered polygons at the
       specified distances for the input geometry array. Options are
       available to union buffers and to use geodesic distance.

       Inputs:
         geometries - The array of geometries to be buffered.
         in_sr - The well-known ID of the spatial reference or a spatial
          reference JSON object for the input geometries.
         distances - The distances that each of the input geometries is
          buffered.
         unit - The units for calculating each buffer distance. If unit
          is not specified, the units are derived from bufferSR. If
          bufferSR is not specified, the units are derived from in_sr.
         out_sr - The well-known ID of the spatial reference or a
          spatial reference JSON object for the input geometries.
         buffer_sr - The well-known ID of the spatial reference or a
          spatial reference JSON object for the input geometries.
         union_results -  If true, all geometries buffered at a given
          distance are unioned into a single (gis,possibly multipart)
          polygon, and the unioned geometry is placed in the output
          array. The default is false
         geodesic - Set geodesic to true to buffer the input geometries
          using geodesic distance. Geodesic distance is the shortest
          path between two points along the ellipsoid of the earth. If
          geodesic is set to false, the 2D Euclidean distance is used
          to buffer the input geometries. The default value depends on
          the geometry type, unit and bufferSR.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.buffer(
        geometries,
        in_sr,
        distances,
        unit,
        out_sr,
        buffer_sr,
        union_results,
        geodesic)

def convex_hull(geometries,
                spatial_ref=None,
                gis=None):
    """
    The convex_hull function is performed on a geometry service
    resource. It returns the convex hull of the input geometry. The
    input geometry can be a point, multipoint, polyline, or polygon.
    The convex hull is typically a polygon but can also be a polyline
    or point in degenerate cases.

    Inputs:
       geometries - The geometries whose convex hull is to be created.
       spatial_ref - The well-known ID or a spatial reference JSON object for
            the output geometry.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.convex_hull(
        geometries,
        spatial_ref)

def cut(cutter,
        target,
        spatial_ref=None,
        gis=None):
    """
    The cut function is performed on a geometry service resource. This
    function splits the target polyline or polygon where it's crossed
    by the cutter polyline.
    At 10.1 and later, this function calls simplify on the input
    cutter and target geometries.

    Inputs:
       cutter - The polyline that will be used to divide the target
        into pieces where it crosses the target.The spatial reference
        of the polylines is specified by spatial_ref. The structure of the
        polyline is the same as the structure of the JSON polyline
        objects returned by the ArcGIS REST API.
       target - The array of polylines/polygons to be cut. The
        structure of the geometry is the same as the structure of the
        JSON geometry objects returned by the ArcGIS REST API. The
        spatial reference of the target geometry array is specified by
        spatial_ref.
       spatial_ref - The well-known ID or a spatial reference JSON object for
        the output geometry.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.cut(
        cutter,
        target,
        spatial_ref)

def densify(geometries,
            spatial_ref,
            max_segment_length,
            length_unit,
            geodesic=False,
            gis=None):
    """
    The densify function is performed using the GIS's geometry engine.
    This function densifies geometries by plotting points between
    existing vertices.

    Inputs:
       geometries - The array of geometries to be densified. The
        structure of each geometry in the array is the same as the
        structure of the JSON geometry objects returned by the ArcGIS
        REST API.
       spatial_ref - The well-known ID or a spatial reference JSON object for
        the input polylines. For a list of valid WKID values, see
        Projected coordinate systems and Geographic coordinate systems.
       max_segment_length - All segments longer than maxSegmentLength are
        replaced with sequences of lines no longer than
        max_segment_length.
       length_unit - The length unit of max_segment_length. If geodesic is
        set to false, then the units are derived from spatial_ref, and
        length_unit is ignored. If geodesic is set to true, then
        length_unit must be a linear unit. In a case where length_unit is
        not specified and spatial_ref is a PCS, the units are derived from spatial_ref.
        In a case where length_unit is not specified and spatial_ref is a GCS,
        then the units are meters.
       geodesic - If geodesic is set to true, then geodesic distance is
        used to calculate max_segment_length. Geodesic distance is the
        shortest path between two points along the ellipsoid of the
        earth. If geodesic is set to false, then 2D Euclidean distance
        is used to calculate max_segment_length. The default is false.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.densify(
        geometries,
        spatial_ref,
        max_segment_length,
        length_unit,
        geodesic)

def difference(geometries,
               spatial_ref,
               geometry,
               gis=None):
    """
    The difference function is performed on a geometry service
    resource. This function constructs the set-theoretic difference
    between each element of an array of geometries and another geometry
    the so-called difference geometry. In other words, let B be the
    difference geometry. For each geometry, A, in the input geometry
    array, it constructs A-B.

    Inputs:
      geometries -  An array of points, multipoints, polylines or
       polygons. The structure of each geometry in the array is the
       same as the structure of the JSON geometry objects returned by
       the ArcGIS REST API.
      geometry - A single geometry of any type and of a dimension equal
       to or greater than the elements of geometries. The structure of
       geometry is the same as the structure of the JSON geometry
       objects returned by the ArcGIS REST API. The use of simple
       syntax is not supported.
      spatial_ref - The well-known ID of the spatial reference or a spatial
       reference JSON object for the input geometries.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.difference(
        geometries,
        spatial_ref,
        geometry)

def distance(spatial_ref,
             geometry1,
             geometry2,
             distance_unit="",
             geodesic=False,
             gis=None):
    """
    The distance function is performed on a geometry service resource.
    It reports the 2D Euclidean or geodesic distance between the two
    geometries.

    Inputs:
     spatial_ref - The well-known ID or a spatial reference JSON object for
      input geometries.
     geometry1 - The geometry from which the distance is to be
      measured. The structure of the geometry is same as the structure
      of the JSON geometry objects returned by the ArcGIS REST API.
     geometry2 - The geometry from which the distance is to be
      measured. The structure of the geometry is same as the structure
      of the JSON geometry objects returned by the ArcGIS REST API.
     distanceUnit - specifies the units for measuring distance between
      the geometry1 and geometry2 geometries.
     geodesic - If geodesic is set to true, then the geodesic distance
      between the geometry1 and geometry2 geometries is returned.
      Geodesic distance is the shortest path between two points along
      the ellipsoid of the earth. If geodesic is set to false or not
      specified, the planar distance is returned. The default value is
      false.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.distance(
        spatial_ref,
        geometry1,
        geometry2,
        distance_unit,
        geodesic)

def find_transformation(in_sr, out_sr, extent_of_interest=None, num_of_results=1, gis=None):
    """
    The find_transformations function is performed on a geometry
    service resource. This function returns a list of applicable
    geographic transformations you should use when projecting
    geometries from the input spatial reference to the output spatial
    reference. The transformations are in JSON format and are returned
    in order of most applicable to least applicable. Recall that a
    geographic transformation is not needed when the input and output
    spatial references have the same underlying geographic coordinate
    systems. In this case, findTransformations returns an empty list.
    Every returned geographic transformation is a forward
    transformation meaning that it can be used as-is to project from
    the input spatial reference to the output spatial reference. In the
    case where a predefined transformation needs to be applied in the
    reverse direction, it is returned as a forward composite
    transformation containing one transformation and a transformForward
    element with a value of false.

    Inputs:
       in_sr - The well-known ID (gis,WKID) of the spatial reference or a
         spatial reference JSON object for the input geometries
       out_sr - The well-known ID (gis,WKID) of the spatial reference or a
         spatial reference JSON object for the input geometries
       extent_of_interest -  The bounding box of the area of interest
         specified as a JSON envelope. If provided, the extent of
         interest is used to return the most applicable geographic
         transformations for the area. If a spatial reference is not
         included in the JSON envelope, the in_sr is used for the
         envelope.
       num_of_results - The number of geographic transformations to
         return. The default value is 1. If num_of_results has a value of
         -1, all applicable transformations are returned.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.find_transformation(in_sr, out_sr,
                                                    extent_of_interest, num_of_results)


def from_geo_coordinate_string(spatial_ref, strings,
                               conversion_type, conversion_mode=None, gis=None):
    """
    The from_geo_coordinate_string function is performed on a geometry
    service resource. The function converts an array of well-known
    strings into xy-coordinates based on the conversion type and
    spatial reference supplied by the user. An optional conversion mode
    parameter is available for some conversion types.

    Inputs:
     spatial_ref - The well-known ID of the spatial reference or a spatial
      reference json object.
     strings - An array of strings formatted as specified by
      conversion_type.
      Syntax: [<string1>,...,<stringN>]
      Example: ["01N AA 66021 00000","11S NT 00000 62155",
                "31U BT 94071 65288"]
     conversion_type - The conversion type of the input strings.
      Valid conversion types are:
       MGRS - Military Grid Reference System
       USNG - United States National Grid
       UTM - Universal Transverse Mercator
       GeoRef - World Geographic Reference System
       GARS - Global Area Reference System
       DMS - Degree Minute Second
       DDM - Degree Decimal Minute
       DD - Decimal Degree
     conversion_mode - Conversion options for MGRS, UTM and GARS
      conversion types.
      Conversion options for MGRS and UTM conversion types.
      Valid conversion modes for MGRS are:
       mgrsDefault - Default. Uses the spheroid from the given spatial
        reference.
       mgrsNewStyle - Treats all spheroids as new, like WGS 1984. The
        180 degree longitude falls into Zone 60.
       mgrsOldStyle - Treats all spheroids as old, like Bessel 1841.
        The 180 degree longitude falls into Zone 60.
       mgrsNewWith180InZone01 - Same as mgrsNewStyle except the 180
        degree longitude falls into Zone 01.
       mgrsOldWith180InZone01 - Same as mgrsOldStyle except the 180
        degree longitude falls into Zone 01.
      Valid conversion modes for UTM are:
       utmDefault - Default. No options.
       utmNorthSouth - Uses north/south latitude indicators instead of
        zone numbers. Non-standard. Default is recommended
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.from_geo_coordinate_string(spatial_ref, strings,
                                                           conversion_type, conversion_mode)


def generalize(spatial_ref,
               geometries,
               max_deviation,
               deviation_unit,
               gis=None):
    """
    The generalize function is performed on a geometry service
    resource. The generalize function simplifies the input geometries
    using the Douglas-Peucker algorithm with a specified maximum
    deviation distance. The output geometries will contain a subset of
    the original input vertices.

    Inputs:
     spatial_ref - The well-known ID or a spatial reference JSON object for the
      input geometries.
     geometries - The array of geometries to be generalized.
     max_deviation - max_deviation sets the maximum allowable offset,
      which will determine the degree of simplification. This value
      limits the distance the output geometry can differ from the input
      geometry.
     deviation_unit - A unit for maximum deviation. If a unit is not
      specified, the units are derived from spatial_ref.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.generalize(
        spatial_ref,
        geometries,
        max_deviation,
        deviation_unit)

def intersect(spatial_ref,
              geometries,
              geometry,
              gis=None):
    """
    The intersect function is performed on a geometry service
    resource. This function constructs the set-theoretic intersection
    between an array of geometries and another geometry. The dimension
    of each resultant geometry is the minimum dimension of the input
    geometry in the geometries array and the other geometry specified
    by the geometry parameter.

    Inputs:
     spatial_ref - The well-known ID or a spatial reference JSON object for the
      input geometries.
     geometries - An array of points, multipoints, polylines, or
      polygons. The structure of each geometry in the array is the same
      as the structure of the JSON geometry objects returned by the
      ArcGIS REST API.
     geometry - A single geometry of any type with a dimension equal to
      or greater than the elements of geometries.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.intersect(spatial_ref,
                                          geometries,
                                          geometry)

def label_points(spatial_ref,
                 polygons,
                 gis=None):
    """
    The label_points function is performed on a geometry service
    resource. The labelPoints function calculates an interior point
    for each polygon specified in the input array. These interior
    points can be used by clients for labeling the polygons.

    Inputs:
     spatial_ref - The well-known ID of the spatial reference or a spatial
      reference JSON object for the input polygons.
     polygons - The array of polygons whose label points are to be
      computed. The spatial reference of the polygons is specified by
      spatial_ref.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.label_points(spatial_ref, polygons)


def lengths(spatial_ref,
            polylines,
            length_unit,
            calculation_type,
            gis=None):
    """
    The lengths function is performed on a geometry service resource.
    This function calculates the 2D Euclidean or geodesic lengths of
    each polyline specified in the input array.

    Inputs:
     spatial_ref - The well-known ID of the spatial reference or a spatial
      reference JSON object for the input polylines.
     polylines - The array of polylines whose lengths are to be
      computed.
     length_unit - The unit in which lengths of polylines will be
      calculated. If calculation_type is planar, then length_unit can be
      any esriUnits constant. If calculation_type is planar and
      length_unit is not specified, then the units are derived from spatial_ref.
      If calculation_type is not planar, then length_unit must be a
      linear esriUnits constant such as esriSRUnit_Meter or
      esriSRUnit_SurveyMile. If calculation_type is not planar and
      length_unit is not specified, then the units are meters.
     calculation_type - calculation_type defines the length calculation
      for the geometry. The type can be one of the following values:
        planar - Planar measurements use 2D Euclidean distance to
         calculate length. This type should only be used if the length
         needs to be calculated in the given spatial reference.
         Otherwise, use preserveShape.
        geodesic - Use this type if you want to calculate a length
         using only the vertices of the polygon and define the lines
         between the vertices as geodesic segments independent of the
         actual shape of the polyline. A geodesic segment is the
         shortest path between two points on an earth ellipsoid.
        preserveShape - This type calculates the length of the geometry
         on the surface of the earth ellipsoid. The shape of the
         geometry in its coordinate system is preserved.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.lengths(
        spatial_ref,
        polylines,
        length_unit,
        calculation_type)

def offset(geometries,
           offset_distance,
           offset_unit,
           offset_how="esriGeometryOffsetRounded",
           bevel_ratio=10,
           simplify_result=False,
           spatial_ref=None,
           gis=None):
    """
    The offset function is performed on a geometry service resource.
    This function constructs geometries that are offset from the
    given input geometries. If the offset parameter is positive, the
    constructed offset will be on the right side of the geometry. Left
    side offsets are constructed with negative parameters. Tracing the
    geometry from its first vertex to the last will give you a
    direction along the geometry. It is to the right and left
    perspective of this direction that the positive and negative
    parameters will dictate where the offset is constructed. In these
    terms, it is simple to infer where the offset of even horizontal
    geometries will be constructed.

    Inputs:
     geometries -  The array of geometries to be offset.
     offset_distance - Specifies the distance for constructing an offset
      based on the input geometries. If the offset_distance parameter is
      positive, the constructed offset will be on the right side of the
      curve. Left-side offsets are constructed with negative values.
     offset_unit - A unit for offset distance. If a unit is not
      specified, the units are derived from spatial_ref.
     offset_how - The offset_how parameter determines how outer corners
      between segments are handled. The three options are as follows:
       esriGeometryOffsetRounded - Rounds the corner between extended
        offsets.
       esriGeometryOffsetBevelled - Squares off the corner after a
        given ratio distance.
       esriGeometryOffsetMitered - Attempts to allow extended offsets
        to naturally intersect, but if that intersection occurs too far
        from the corner, the corner is eventually bevelled off at a
        fixed distance.
     bevel_ratio - bevel_ratio is multiplied by the offset distance, and
      the result determines how far a mitered offset intersection can
      be located before it is bevelled. When mitered is specified,
      bevel_ratio is ignored and 10 is used internally. When bevelled is
      specified, 1.1 will be used if bevel_ratio is not specified.
      bevel_ratio is ignored for rounded offset.
     simplify_result - if simplify_result is set to true, then self
      intersecting loops will be removed from the result offset
      geometries. The default is false.
     spatial_ref - The well-known ID or a spatial reference JSON object for the
      input geometries.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.offset(
        geometries,
        offset_distance,
        offset_unit,
        offset_how,
        bevel_ratio,
        simplify_result,
        spatial_ref)


def project(geometries,
            in_sr,
            out_sr,
            transformation="",
            transform_forward=False,
            gis=None):
    """
    The project function is performed on a geometry service resource.
    This function projects an array of input geometries from the input
    spatial reference to the output spatial reference.

    Inputs:
     geometries - The list of geometries to be projected.
     in_sr - The well-known ID (gis,WKID) of the spatial reference or a
      spatial reference JSON object for the input geometries.
     out_sr - The well-known ID (gis,WKID) of the spatial reference or a
      spatial reference JSON object for the input geometries.
     transformation - The WKID or a JSON object specifying the
      geographic transformation (gis,also known as datum transformation) to
      be applied to the projected geometries. Note that a
      transformation is needed only if the output spatial reference
      contains a different geographic coordinate system than the input
      spatial reference.
     transform_forward - A Boolean value indicating whether or not to
      transform forward. The forward or reverse direction of
      transformation is implied in the name of the transformation. If
      transformation is specified, a value for the transformForward
      parameter must also be specified. The default value is false.

    Example:
     input_geom = [{"x": -17568824.55, "y": 2428377.35}, {"x": -17568456.88, "y": 2428431.352}]
     result = project(geometries = input_geom, in_sr = 3857, out_sr = 4326)

    returns:
     a list of geometries in the out_sr coordinate system, for instance:
     [{"x": -157.82343617279275, "y": 21.305781607280093}, {"x": -157.8201333369876, "y": 21.306233559873714}]
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.project(
        geometries,
        in_sr,
        out_sr,
        transformation,
        transform_forward)


def relation(geometries1,
             geometries2,
             spatial_ref,
             spatial_relation="esriGeometryRelationIntersection",
             relation_param="",
             gis=None):
    """
    The relation function is performed on a geometry service resource.
    This function determines the pairs of geometries from the input
    geometry arrays that participate in the specified spatial relation.
    Both arrays are assumed to be in the spatial reference specified by
    spatial_ref, which is a required parameter. Geometry types cannot be mixed
    within an array. The relations are evaluated in 2D. In other words,
    z coordinates are not used.

    Inputs:
     geometries1 - The first array of geometries used to compute the
      relations.
     geometries2 -The second array of geometries used to compute the
     relations.
     spatial_ref - The well-known ID of the spatial reference or a spatial
      reference JSON object for the input geometries.
     spatial_relation - The spatial relationship to be tested between the two
      input geometry arrays.
      Values: esriGeometryRelationCross | esriGeometryRelationDisjoint |
      esriGeometryRelationIn | esriGeometryRelationInteriorIntersection |
      esriGeometryRelationIntersection | esriGeometryRelationLineCoincidence |
      esriGeometryRelationLineTouch | esriGeometryRelationOverlap |
      esriGeometryRelationPointTouch | esriGeometryRelationTouch |
      esriGeometryRelationWithin | esriGeometryRelationRelation
     relation_param - The Shape Comparison Language string to be
      evaluated.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.relation(
        geometries1,
        geometries2,
        spatial_ref,
        spatial_relation,
        relation_param)


def reshape(spatial_ref,
            target,
            reshaper,
            gis=None):
    """
    The reshape function is performed on a geometry service resource.
    It reshapes a polyline or polygon feature by constructing a
    polyline over the feature. The feature takes the shape of the
    reshaper polyline from the first place the reshaper intersects the
    feature to the last.

    Input:
     spatial_ref - The well-known ID of the spatial reference or a spatial
      reference JSON object for the input geometries.
     target -  The polyline or polygon to be reshaped.
     reshaper - The single-part polyline that does the reshaping.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.reshape(
        spatial_ref,
        target,
        reshaper)


def simplify(spatial_ref,
             geometries,
             gis=None):
    """
    The simplify function is performed on a geometry service resource.
    Simplify permanently alters the input geometry so that the geometry
    becomes topologically consistent. This resource applies the ArcGIS
    simplify function to each geometry in the input array.

    Inputs:
    spatial_ref - The well-known ID of the spatial reference or a spatial
      reference JSON object for the input geometries.
    geometries - The array of geometries to be simplified.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.simplify(spatial_ref, geometries)


def to_geo_coordinate_string(spatial_ref,
                             coordinates,
                             conversion_type,
                             conversion_mode="mgrsDefault",
                             num_of_digits=None,
                             rounding=True,
                             add_spaces=True,
                             gis=None):
    """
    The to_geo_coordinate_string function is performed on a geometry
    service resource. The function converts an array of
    xy-coordinates into well-known strings based on the conversion type
    and spatial reference supplied by the user. Optional parameters are
    available for some conversion types. Note that if an optional
    parameter is not applicable for a particular conversion type, but a
    value is supplied for that parameter, the value will be ignored.

    Inputs:
      spatial_ref -  The well-known ID of the spatial reference or a spatial
       reference json object.
      coordinates - An array of xy-coordinates in JSON format to be
       converted. Syntax: [[x1,y2],...[xN,yN]]
      conversion_type - The conversion type of the input strings.
       Allowed Values:
        MGRS - Military Grid Reference System
        USNG - United States National Grid
        UTM - Universal Transverse Mercator
        GeoRef - World Geographic Reference System
        GARS - Global Area Reference System
        DMS - Degree Minute Second
        DDM - Degree Decimal Minute
        DD - Decimal Degree
      conversion_mode - Conversion options for MGRS and UTM conversion
       types.
       Valid conversion modes for MGRS are:
        mgrsDefault - Default. Uses the spheroid from the given spatial
         reference.
        mgrsNewStyle - Treats all spheroids as new, like WGS 1984. The
         180 degree longitude falls into Zone 60.
        mgrsOldStyle - Treats all spheroids as old, like Bessel 1841.
         The 180 degree longitude falls into Zone 60.
        mgrsNewWith180InZone01 - Same as mgrsNewStyle except the 180
         degree longitude falls into Zone 01.
        mgrsOldWith180InZone01 - Same as mgrsOldStyle except the 180
         degree longitude falls into Zone 01.
       Valid conversion modes for UTM are:
        utmDefault - Default. No options.
        utmNorthSouth - Uses north/south latitude indicators instead of
         zone numbers. Non-standard. Default is recommended.
      num_of_digits - The number of digits to output for each of the
       numerical portions in the string. The default value for
       num_of_digits varies depending on conversion_type.
      rounding - If true, then numeric portions of the string are
       rounded to the nearest whole magnitude as specified by
       numOfDigits. Otherwise, numeric portions of the string are
       truncated. The rounding parameter applies only to conversion
       types MGRS, USNG and GeoRef. The default value is true.
      addSpaces - If true, then spaces are added between components of
       the string. The addSpaces parameter applies only to conversion
       types MGRS, USNG and UTM. The default value for MGRS is false,
       while the default value for both USNG and UTM is true.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.to_geo_coordinate_string(
        spatial_ref,
        coordinates,
        conversion_type,
        conversion_mode,
        num_of_digits,
        rounding,
        add_spaces)


def trim_extend(spatial_ref,
                polylines,
                trim_extend_to,
                extend_how=0,
                gis=None):
    """
    The trim_extend function is performed on a geometry service
    resource. This function trims or extends each polyline specified
    in the input array, using the user-specified guide polylines. When
    trimming features, the part to the left of the oriented cutting
    line is preserved in the output, and the other part is discarded.
    An empty polyline is added to the output array if the corresponding
    input polyline is neither cut nor extended.

    Inputs:
     spatial_ref - The well-known ID of the spatial reference or a spatial
       reference json object.
     polylines - An array of polylines to be trimmed or extended.
     trim_extend_to - A polyline that is used as a guide for trimming or
      extending input polylines.
     extend_how - A flag that is used along with the trimExtend
      function.
      0 - By default, an extension considers both ends of a path. The
       old ends remain, and new points are added to the extended ends.
       The new points have attributes that are extrapolated from
       adjacent existing segments.
      1 - If an extension is performed at an end, relocate the end
       point to the new position instead of leaving the old point and
       adding a new point at the new position.
      2 - If an extension is performed at an end, do not extrapolate
       the end-segment's attributes for the new point. Instead, make
       its attributes the same as the current end. Incompatible with
       esriNoAttributes.
      4 - If an extension is performed at an end, do not extrapolate
       the end-segment's attributes for the new point. Instead, make
       its attributes empty. Incompatible with esriKeepAttributes.
      8 - Do not extend the 'from' end of any path.
      16 - Do not extend the 'to' end of any path.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.trim_extend(spatial_ref,
                                            polylines,
                                            trim_extend_to,
                                            extend_how)


def union(spatial_ref,
          geometries,
          gis=None):
    """
    The union function is performed on a geometry service resource.
    This function constructs the set-theoretic union of the geometries
    in the input array. All inputs must be of the same type.

    Inputs:
    spatial_ref - The well-known ID of the spatial reference or a spatial
     reference json object.
    geometries - The array of geometries to be unioned.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    return gis._tools.geometry.union(spatial_ref, geometries)



""" The store module defines functions for interacting with the Workforce backend.
"""
from . import assignments
from . import assignment_types
from . import attachments
from . import dispatchers
from . import workers
from . import tracks
from . import projects

from .assignment_types import (
    get_assignment_type,
    get_assignment_types,
    add_assignment_types,
    update_assignment_types,
    delete_assignment_types,
    add_assignment_type,
    update_assignment_type
)

from .assignments import (
    get_assignment,
    get_assignments,
    query_assignments,
    add_assignments,
    update_assignments,
    delete_assignments,
    add_assignment,
    update_assignment
)

from .attachments import (
    get_attachments,
    add_attachment,
    delete_attachments
)

from .dispatchers import (
    get_dispatcher,
    get_dispatchers,
    query_dispatchers,
    add_dispatchers,
    update_dispatchers,
    delete_dispatchers,
    add_dispatcher,
    update_dispatcher
)

from .projects import (
    get_project
)

from .tracks import (
    get_track,
    get_tracks,
    query_tracks,
    delete_tracks,
    add_tracks,
    add_track,
    update_tracks,
    update_track
)

from .workers import (
    get_worker,
    get_workers,
    query_workers,
    add_workers,
    update_workers,
    delete_workers,
    add_worker,
    update_worker
)



from . import workforce
from . import storymap
from . import survey123
from . _url_schemes import build_collector_url
from . _url_schemes import build_explorer_url
from . _url_schemes import build_navigator_url
from . _url_schemes import build_survey123_url



"""
Contains the base class that all server object inherit from.
"""
from __future__ import absolute_import
import json
from collections import OrderedDict
from ._connection import ServerConnection
from urllib.request import HTTPError
from arcgis._impl.connection import _ArcGISConnection
from arcgis.gis import GIS
from arcgis._impl.common._mixins import PropertyMap
###########################################################################
class BaseServer(object):
    """class most server object inherit from"""
    _con = None
    _url = None
    _json_dict = None
    _json = None
    _properties = None
    def __init__(self, url, gis=None, initialize=True, **kwargs):
        """class initializer"""
        if gis is None and 'connection' in kwargs:
            connection = kwargs['connection']
            gis = kwargs.pop('connection', None)
        super(BaseServer, self).__init__()
        self._url = url

        #gis = kwargs.pop('gis', None)
        if gis is None and \
           isinstance(gis, GIS):
            gis = gis._portal.con

        if isinstance(gis, (ServerConnection, _ArcGISConnection)):
            self._con = gis
        elif hasattr(gis, '_con'):
            self._gis = gis._con
        else:
            raise ValueError("gis must be of type SiteConnection")
        if initialize:
            self._init(gis)
    #----------------------------------------------------------------------
    def _init(self, connection=None):
        """loads the properties into the class"""
        if connection is None:
            connection = self._con
        params = {"f":"json"}
        try:
            result = connection.get(path=self._url,
                                    params=params)
            if isinstance(result, dict):
                self._json_dict = result
                self._properties = PropertyMap(result)
            else:
                self._json_dict = {}
                self._properties = PropertyMap({})
        except HTTPError as err:
            raise RuntimeError(err)
        except:
            self._json_dict = {}
            self._properties = PropertyMap({})
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self._url)
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """
        returns the object properties
        """
        if self._properties is None:
            self._init()
        return self._properties
    #----------------------------------------------------------------------
    def __getattr__(self, name):
        """adds dot notation to any class"""
        if self._properties is None:
            self._init()
        try:
            return self._properties.__getitem__(name)
        except:
            for k,v in self._json_dict.items():
                if k.lower() == name.lower():
                    return v
            raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__, name))
    #----------------------------------------------------------------------
    def __getitem__(self, key):
        """helps make object function like a dictionary object"""
        try:
            return self._properties.__getitem__(key)
        except KeyError:
            for k,v in self._json_dict.items():
                if k.lower() == key.lower():
                    return v
            raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__,
                                                                        key))
        except:
            raise AttributeError("'%s' object has no attribute '%s'" % (type(self).__name__,
                                                                        key))
    #----------------------------------------------------------------------
    @property
    def url(self):
        """gets/sets the service url"""
        return self._url
    #----------------------------------------------------------------------
    @url.setter
    def url(self, value):
        """gets/sets the service url"""
        self._url = value
        self.refresh()
    #----------------------------------------------------------------------
    def __iter__(self):
        """creates iterable for classes properties"""
        for k,v in self._json_dict.items():
            yield k,v
    #----------------------------------------------------------------------
    def _refresh(self):
        """reloads all the properties of a given service"""
        self._init()
